# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utils for managing resources like cluster, VM and host."""

from oslo_log import log

from networking_vsphere.utils import cache
from networking_vsphere.utils import common_util
from networking_vsphere.utils import vim_util

LOG = log.getLogger(__name__)


def get_host_mor_for_vm(session, vm_uuid):
    """Return host mor from VM uuid."""

    vm_mor = get_vm_mor_for_uuid(session, vm_uuid)
    if vm_mor:
        host_mor = session._call_method(
            vim_util, "get_dynamic_property", vm_mor,
            "VirtualMachine", "runtime.host")
        return host_mor
    return None


def get_hostname_for_host_mor(session, host_mor):
    """Return hostname from host mor."""

    if host_mor:
        hostname = session._call_method(
            vim_util, "get_dynamic_property", host_mor,
            "HostSystem", "name")
        return hostname
    return None


def get_host_mor_by_name(session, host_name):
    """Return Host mor from its name."""

    host_mors = session._call_method(
        vim_util, "get_objects", "HostSystem", ["name"])
    for host_mor in host_mors:
        propset_dict = common_util.convert_propset_to_dict(host_mor.propSet)
        if propset_dict['name'] == host_name:
            return host_mor.obj
    return None


def get_vm_mor_by_name(session, vm_name):
    """Return VM mor from its name."""

    vm_mors = session._call_method(
        vim_util, "get_objects", "VirtualMachine", ["name"])
    for vm_mor in vm_mors:
        propset_dict = common_util.convert_propset_to_dict(vm_mor.propSet)
        if propset_dict['name'] == vm_name:
            return vm_mor.obj
    return None


def set_vm_poweroff(session, vm_mor):
    """Power off the VM."""

    try:
        task_ref = session._call_method(
            session._get_vim(), "PowerOffVM_Task", vm_mor)
        session.wait_for_task(task_ref)
    except Exception as e:
        LOG.exception(_("%s"), e)
        raise Exception(e)


def set_host_into_maintenance_mode(session, host_mor):
    """Put ESX host into maintenance mode."""

    try:
        task_ref = session._call_method(
            session._get_vim(), "EnterMaintenanceMode_Task", host_mor,
            timeout=0, evacuatePoweredOffVms=False)
        session.wait_for_task(task_ref)
    except Exception as e:
        LOG.exception(_("%s"), e)
        raise Exception(e)


def set_host_into_shutdown_mode(session, host_mor):
    """Shutdown the ESX host."""

    try:
        enabled = _check_shutdown_enabled(session, host_mor)
        if enabled:
            task_ref = session._call_method(
                session._get_vim(), "ShutdownHost_Task", host_mor,
                force=True)
            session.wait_for_task(task_ref)
    except Exception as e:
        LOG.exception(_("%s"), e)
        raise Exception(e)


def _check_shutdown_enabled(session, host_mor):
    """Check shutdown enabled on ESX host."""

    if host_mor:
        shutdown_enabled = session._call_method(
            vim_util, "get_dynamic_property", host_mor,
            "HostSystem", "capability.shutdownSupported")
        return shutdown_enabled
    return False


def get_clustername_for_cluster_mor(session, cluster_mor):
    """Return cluster name from cluster mor."""

    if cluster_mor:
        clustername = session._call_method(
            vim_util, "get_dynamic_property", cluster_mor,
            "ClusterComputeResource", "name")
        return clustername
    return None


def get_clusterid_for_cluster_mor(session, cluster_mor):
    """Return cluster id from cluster mor."""

    if cluster_mor:
        clusterid = cluster_mor.value
        return clusterid
    return None


def get_cluster_mor_for_vm(session, vm_uuid):
    """Return cluster mor from VM uuid."""

    cluster_mor = cache.VCCache.get_cluster_mor_for_vm(vm_uuid)
    if cluster_mor:
        return cluster_mor
    vm_mor = get_vm_mor_for_uuid(session, vm_uuid)
    if vm_mor:
        host_mor = session._call_method(
            vim_util, "get_dynamic_property", vm_mor,
            "VirtualMachine", "runtime.host")
        cluster_mor = session._call_method(
            vim_util, "get_dynamic_property", host_mor,
            "HostSystem", "parent")
        cache.VCCache.add_cluster_mor_for_vm(vm_uuid, cluster_mor)
        return cluster_mor
    LOG.debug("Unable to retrieve VM information")
    return None


def get_vm_mor_for_uuid(session, vm_uuid):
    """Return VM mor from VM uuid."""

    vm_mor = cache.VCCache.get_vm_mor_for_uuid(vm_uuid)
    if vm_mor:
        return vm_mor
    # TODO(romilg) : use config.uuid instead of
    # config.extraConfig["nvp.vm-uuid"] once fixed in nova.
    vm_mors = session._call_method(vim_util, "get_objects",
                                   "VirtualMachine",
                                   ['config.extraConfig["nvp.vm-uuid"]'])
    for vm_mor in vm_mors:
        if hasattr(vm_mor, "propSet"):
            propset_dict = common_util.convert_propset_to_dict(vm_mor.propSet)
            if (vm_uuid ==
                    propset_dict['config.extraConfig["nvp.vm-uuid"]'].value):
                cache.VCCache.add_vm_mor_for_uuid(vm_uuid, vm_mor.obj)
                return vm_mor.obj
    LOG.debug("Unable to retrieve VM information")
    return None


def get_host_mors_for_cluster(session, cluster_mor):
    """Return host mors from cluster mor."""

    try:
        host_mors = None
        host_ret = session._call_method(vim_util, "get_dynamic_property",
                                        cluster_mor,
                                        "ClusterComputeResource", "host")
        if hasattr(host_ret, "ManagedObjectReference"):
            host_mors = host_ret.ManagedObjectReference
        return host_mors
    except Exception as e:
        LOG.exception(_("Error retrieving cluster information"))
        raise Exception(e)


def get_extraconfigs_for_vm(session, vm_mor):
    """Return extra configurations from VM mor."""

    optvals = session._call_method(vim_util,
                                   "get_dynamic_property",
                                   vm_mor,
                                   "VirtualMachine",
                                   "config.extraConfig")
    optvals = optvals.OptionValue
    extraconfigs = {}
    for optval in optvals:
        extraconfigs[optval.key] = optval.value
    return extraconfigs


def get_cluster_mor_by_path(session, path):
    """Return cluster mor from cluster path."""

    search_index = session._call_method(vim_util, "get_search_index")
    mor = session._call_method(
        vim_util, "find_by_inventory_path", search_index, path)
    return mor


def _get_token(results):
    """Get the token from the property results."""
    return getattr(results, 'token', None)


def _get_object_for_value(results, value):
    """Get vm reference value from vm object."""

    for object in results:
        if object.propSet[0].val == value:
            return object.obj


def _get_object_from_results(session, results, value, func):
    """Get a specific vm object from list of vm objects."""

    while results:
        token = _get_token(results)
        object = func(results, value)

        if object:
            if token:
                session._call_method(vim_util,
                                     "cancel_retrieve",
                                     token)
            return object

        if token:
            results = session._call_method(vim_util,
                                           "continue_to_get_objects",
                                           token)
        else:
            return None


def get_vm_reference(session, vm_uuid):
    """Get vm reference from uuid."""

    vms = session._call_method(vim_util, "get_objects",
                               "VirtualMachine", ["name"])
    return _get_object_from_results(session, vms, vm_uuid,
                                    _get_object_for_value)



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import sys
import traceback

from oslo_log import log

from networking_vsphere.common import error

LOG = log.getLogger(__name__)


def import_class(import_str):
    """Returns a class from a string including module and class."""
    mod_str, _sep, class_str = import_str.rpartition('.')
    try:
        __import__(mod_str)
        return getattr(sys.modules[mod_str], class_str)
    except (ValueError, AttributeError):
        raise ImportError('Class %s cannot be found (%s).' %
                          (class_str,
                           traceback.format_exception(*sys.exc_info())))


def load_object(driver, base_class, *args, **kwargs):
    """Load a class, instantiate, check if its of base_class type."""
    driver_obj = import_class(driver)(*args, **kwargs)
    if not isinstance(driver_obj, base_class):
        raise TypeError("Invalid type - %s does not extend %s." %
                        (fullname(driver_obj), base_class))
    return driver_obj


def fullname(cls):
    """Get full name of a class."""
    module = cls.__module__
    if module is None or module == str.__class__.__module__:
        return cls.__class__.__name__
    return module + '.' + cls.__class__.__name__


def require_state(state=None, excp=True):
    """Decorator to check state of an object.

    First argument of the decorated function should be
    the object whose state needs to be checked.
    :param state: valid set of states
    :param excp: If True then raise an exception if in invalid state
    """
    if state is not None and not isinstance(state, set):
        state = set(state)

    def outer(f):
        @functools.wraps(f)
        def inner(obj, *args, **kw):
            if state is not None and obj.state not in state:
                l_states = list(state)
                if excp:
                    raise error.OVSvAppNeutronAgentError(
                        "%s not allowed. "
                        "%s is in %s state. "
                        "To be in %s state" %
                        (f.__name__,
                         obj.__class__.__name__,
                         obj.state,
                         l_states))
                else:
                    LOG.info(_("%(name)s not allowed. "
                               "%(obj)s is %(state)s state. "
                               "Need to be in %(states)s state."),
                             {'name': f.__name__,
                              'obj': obj.__class__.__name__,
                              'state': obj.state,
                              'states': l_states})
                    return
            return f(obj, *args, **kw)
        return inner
    return outer



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

import mock

import contextlib

from oslo_config import cfg

from networking_vsphere.agent import ovsvapp_agent
from networking_vsphere.common import constants as ovsvapp_const
from networking_vsphere.common import error
from networking_vsphere.tests import base
from networking_vsphere.tests.unit.drivers import fake_manager

from neutron.agent.common import ovs_lib
from neutron.plugins.common import constants as p_const

VNIC_ADDED = 'VNIC_ADDED'
FAKE_DEVICE_ID = 'fake_device_id'
FAKE_VM = 'fake_vm'
FAKE_HOST_1 = 'fake_host_1'
FAKE_HOST_2 = 'fake_host_2'
FAKE_CLUSTER_MOID = 'fake_cluster_moid'
FAKE_CLUSTER_1 = 'fake_cluster_1'
FAKE_CLUSTER_2 = 'fake_cluster_2'
FAKE_VCENTER = 'fake_vcenter'
FAKE_PORT_1 = 'fake_port_1'
FAKE_PORT_2 = 'fake_port_2'
MAC_ADDRESS = '01:02:03:04:05:06'
FAKE_CONTEXT = 'fake_context'
FAKE_SG = {'fake_sg': 'fake_sg_rule'}
FAKE_SG_RULES = {FAKE_DEVICE_ID: ['fake_rule_1',
                                  'fake_rule_2',
                                  'fake_rule_3']
                 }
DEVICE = {'id': FAKE_DEVICE_ID,
          'cluster_id': FAKE_CLUSTER_1,
          'host': FAKE_HOST_1,
          'vcenter': FAKE_VCENTER}


class SampleEvent(object):
    def __init__(self, type, host, cluster, srcobj, host_changed=False):
        self.event_type = type
        self.host_name = host
        self.cluster_id = cluster
        self.src_obj = srcobj
        self.host_changed = host_changed


class VM(object):
    def __init__(self, uuid, vnics):
        self.uuid = uuid
        self.vnics = vnics


class SamplePort(object):
    def __init__(self, port_uuid):
        self.port_uuid = port_uuid


class SamplePortUIDMac(object):
    def __init__(self, port_uuid, mac_address):
        self.port_uuid = port_uuid
        self.mac_address = mac_address


class TestOVSvAppL2Agent(base.TestCase):

    def setUp(self):
        super(TestOVSvAppL2Agent, self).setUp()
        cfg.CONF.set_override('security_bridge_mapping',
                              "br-fake:fake_if", 'SECURITYGROUP')
        with contextlib.nested(
            mock.patch('neutron.common.config.'
                       'init'),
            mock.patch('neutron.common.config.'
                       'setup_logging'),
            mock.patch('networking_vsphere.agent.ovsvapp_agent.'
                       'RpcPluginApi'),
            mock.patch('neutron.agent.securitygroups_rpc.'
                       'SecurityGroupServerRpcApi'),
            mock.patch('neutron.agent.rpc.'
                       'PluginReportStateAPI'),
            mock.patch('networking_vsphere.agent.ovsvapp_agent.'
                       'OVSvAppPluginApi'),
            mock.patch('neutron.context.'
                       'get_admin_context_without_session'),
            mock.patch('neutron.agent.rpc.'
                       'create_consumers'),
            mock.patch('neutron.plugins.openvswitch.agent.ovs_neutron_agent.'
                       'OVSNeutronAgent.setup_integration_br'),
            mock.patch('networking_vsphere.agent.ovsvapp_agent.'
                       'OVSvAppL2Agent.setup_ovs_bridges'),
            mock.patch('networking_vsphere.agent.ovsvapp_agent.'
                       'OVSvAppL2Agent.setup_security_br'),
            mock.patch('networking_vsphere.agent.ovsvapp_agent.'
                       'OVSvAppL2Agent._init_ovs_flows'),
            mock.patch('networking_vsphere.drivers.'
                       'ovs_firewall.OVSFirewallDriver.'
                       'setup_base_flows'),
            mock.patch('neutron.agent.common.ovs_lib.OVSBridge.'
                       'create'),
            mock.patch('neutron.agent.common.ovs_lib.OVSBridge.'
                       'set_secure_mode'),
            mock.patch('neutron.agent.common.ovs_lib.OVSBridge.'
                       'get_port_ofport',
                       return_value=5)):
            self.agent = ovsvapp_agent.OVSvAppL2Agent()
            self.agent.run_refresh_firewall_loop = False
        self.LOG = ovsvapp_agent.LOG

    def _build_port(self):
        port = {'admin_state_up': False,
                'id': FAKE_PORT_1,
                'device': DEVICE,
                'network_id': 'net_uuid',
                'physical_network': 'physnet1',
                'segmentation_id': '1001',
                'lvid': 1,
                'network_type': 'vlan',
                'fixed_ips': [{'subnet_id': 'subnet_uuid',
                               'ip_address': '1.1.1.1'}],
                'device_owner': 'compute:None',
                'security_groups': FAKE_SG,
                'mac_address': MAC_ADDRESS,
                'device_id': FAKE_DEVICE_ID,
                }
        return port

    def test_setup_security_br_none(self):
        cfg.CONF.set_override('security_bridge_mapping',
                              None, 'SECURITYGROUP')
        self.agent.sec_br = mock.Mock()
        with contextlib.nested(
            mock.patch.object(self.LOG, 'warn'),
            mock.patch.object(self.agent.sec_br, 'bridge_exists')
        ) as (logger_warn, ovs_bridge):
            self.assertRaises(SystemExit,
                              self.agent.setup_security_br)
            self.assertTrue(logger_warn.called)
            self.assertFalse(ovs_bridge.called)

    def test_setup_security_br(self):
        cfg.CONF.set_override('security_bridge_mapping',
                              "br-fake:fake_if", 'SECURITYGROUP')
        self.agent.sec_br = mock.Mock()
        self.agent.int_br = mock.Mock()
        with contextlib.nested(
            mock.patch.object(self.LOG, 'info'),
            mock.patch.object(ovs_lib, "OVSBridge"),
            mock.patch.object(self.agent.sec_br,
                              "add_patch_port",
                              return_value=5),
            mock.patch.object(self.agent.int_br,
                              "add_patch_port",
                              return_value=6),
        )as (logger_info, ovs_br, sec_add_patch_port, int_add_patch_port):
            self.agent.setup_security_br()
            self.assertTrue(ovs_br.called)
            self.assertTrue(self.agent.sec_br.add_patch_port.called)
            self.assertTrue(logger_info.called)

    def test_recover_security_br_none(self):
        cfg.CONF.set_override('security_bridge_mapping',
                              None, 'SECURITYGROUP')
        self.agent.sec_br = mock.Mock()
        with contextlib.nested(
            mock.patch.object(self.LOG, 'warn'),
            mock.patch.object(self.agent.sec_br, 'bridge_exists')
        ) as (logger_warn, ovs_bridge):
            self.assertRaises(SystemExit,
                              self.agent.recover_security_br)
            self.assertTrue(logger_warn.called)
            self.assertFalse(ovs_bridge.called)

    def test_update_port_bindings(self):
        self.agent.ports_to_bind.add("fake_port")
        with contextlib.nested(
            mock.patch.object(self.agent.ovsvapp_rpc,
                              "update_ports_binding",
                              return_value=set(["fake_port"])),
            mock.patch.object(self.LOG, 'exception'),
        ) as (update_ports_binding, log_exception):
            self.agent._update_port_bindings()
            self.assertTrue(update_ports_binding.called)
            self.assertFalse(self.agent.ports_to_bind)
            self.assertFalse(log_exception.called)

    def test_update_port_bindings_rpc_exception(self):
        self.agent.ports_to_bind.add("fake_port")
        with contextlib.nested(
            mock.patch.object(self.agent.ovsvapp_rpc,
                              "update_ports_binding",
                              side_effect=Exception()),
            mock.patch.object(self.LOG, 'exception'),
        ) as (update_port_binding, log_exception):
            self.assertRaises(
                error.OVSvAppNeutronAgentError,
                self.agent._update_port_bindings)
            self.assertTrue(update_port_binding.called)
            self.assertTrue(log_exception.called)
            self.assertEqual(set(['fake_port']),
                             self.agent.ports_to_bind)

    def test_update_port_bindings_partial(self):
        self.agent.ports_to_bind.add("fake_port1")
        self.agent.ports_to_bind.add("fake_port2")
        self.agent.ports_to_bind.add("fake_port3")
        with contextlib.nested(
            mock.patch.object(self.agent.ovsvapp_rpc,
                              "update_ports_binding",
                              return_value=set(["fake_port1",
                                                "fake_port2"])),
            mock.patch.object(self.LOG, 'exception'),
        ) as (update_port_binding, log_exception):
            self.agent._update_port_bindings()
            self.assertTrue(update_port_binding.called)
            self.assertEqual(set(["fake_port3"]),
                             self.agent.ports_to_bind)

    def test_setup_ovs_bridges_vlan(self):
        cfg.CONF.set_override('tenant_network_type',
                              "vlan", 'OVSVAPP')
        cfg.CONF.set_override('bridge_mappings',
                              ["physnet1:br-eth1"], 'OVSVAPP')
        with contextlib.nested(
            mock.patch.object(self.agent, 'setup_physical_bridges'),
            mock.patch.object(self.agent, '_init_ovs_flows')
        ) as (mock_phys_brs, mock_init_ovs_flows):
            self.agent.setup_ovs_bridges()
            mock_phys_brs.assert_called_with(self.agent.bridge_mappings)
            mock_init_ovs_flows.assert_called_with(self.agent.bridge_mappings)

    def test_setup_ovs_bridges_vxlan(self):
        cfg.CONF.set_override('tenant_network_type',
                              "vxlan", 'OVSVAPP')
        cfg.CONF.set_override('local_ip',
                              "10.10.10.10", 'OVSVAPP')
        cfg.CONF.set_override('tunnel_bridge',
                              "br-tun", 'OVSVAPP')
        with contextlib.nested(
            mock.patch.object(self.agent, 'reset_tunnel_br'),
            mock.patch.object(self.agent, 'setup_tunnel_br')
            ) as (mock_reset_tunnel_br, mock_setup_tunnel_br):
            self.agent.setup_ovs_bridges()
            mock_reset_tunnel_br.assert_called_with("br-tun")
            self.assertTrue(mock_setup_tunnel_br.called)

    def test_mitigate_ovs_restart_vlan(self):
        self.agent.enable_tunneling = False
        self.agent.refresh_firewall_required = False
        self.agent.devices_to_filter = set(['1111'])
        self.agent.cluster_host_ports = set(['1111'])
        self.agent.cluster_other_ports = set(['2222'])
        with contextlib.nested(
            mock.patch.object(self.LOG, 'info'),
            mock.patch.object(self.agent, "setup_integration_br"),
            mock.patch.object(self.agent, "setup_physical_bridges"),
            mock.patch.object(self.agent, "setup_security_br"),
            mock.patch.object(self.agent.sg_agent, "init_firewall"),
            mock.patch.object(self.agent, "reset_tunnel_br"),
            mock.patch.object(self.agent, "setup_tunnel_br"),
            mock.patch.object(self.agent, "_init_ovs_flows"),
        )as (logger_info, int_br, phys_brs, sec_br, init_fw, reset_tun_br,
             tun_br, init_flows):
            self.agent.mitigate_ovs_restart()
            self.assertTrue(int_br.called)
            self.assertTrue(phys_brs.called)
            self.assertTrue(sec_br.called)
            self.assertFalse(reset_tun_br.called)
            self.assertFalse(tun_br.called)
            self.assertTrue(init_fw.called)
            self.assertTrue(init_flows.called)
            self.assertTrue(logger_info.called)
            self.assertTrue(self.agent.refresh_firewall_required)
            self.assertEqual(2, len(self.agent.devices_to_filter))

    def test_mitigate_ovs_restart_vxlan(self):
        self.agent.enable_tunneling = True
        self.agent.refresh_firewall_required = False
        self.agent.devices_to_filter = set(['1111'])
        self.agent.cluster_host_ports = set(['1111'])
        self.agent.cluster_other_ports = set(['2222'])
        with contextlib.nested(
            mock.patch.object(self.LOG, 'info'),
            mock.patch.object(self.agent, "setup_integration_br"),
            mock.patch.object(self.agent, "setup_physical_bridges"),
            mock.patch.object(self.agent, "setup_security_br"),
            mock.patch.object(self.agent.sg_agent, "init_firewall"),
            mock.patch.object(self.agent, "reset_tunnel_br"),
            mock.patch.object(self.agent, "setup_tunnel_br"),
            mock.patch.object(self.agent, "tunnel_sync"),
            mock.patch.object(self.agent, "_init_ovs_flows"),
        )as (logger_info, int_br, phys_brs, sec_br, init_fw, reset_tun_br,
             tun_br, tun_sync, init_flows):
            self.agent.mitigate_ovs_restart()
            self.assertTrue(reset_tun_br.called)
            self.assertTrue(tun_br.called)
            self.assertFalse(phys_brs.called)
            self.assertTrue(tun_sync.called)
            self.assertTrue(logger_info.called)
            self.assertTrue(self.agent.refresh_firewall_required)
            self.assertEqual(len(self.agent.devices_to_filter), 2)

    def test_mitigate_ovs_restart_exception(self):
        self.agent.enable_tunneling = False
        self.agent.refresh_firewall_required = False
        self.agent.devices_to_filter = set()
        self.agent.cluster_host_ports = set(['1111'])
        self.agent.cluster_other_ports = set(['2222'])

        with contextlib.nested(
            mock.patch.object(self.LOG, "info"),
            mock.patch.object(self.agent, "setup_integration_br",
                              side_effect=Exception()),
            mock.patch.object(self.agent, "setup_physical_bridges"),
            mock.patch.object(self.agent, "reset_tunnel_br"),
            mock.patch.object(self.agent, "setup_tunnel_br"),
            mock.patch.object(self.LOG, "exception"),
        )as (logger_info, int_br, phys_brs, reset_tun_br, tun_br,
             exception_log):
            self.agent.mitigate_ovs_restart()
            self.assertTrue(int_br.called)
            self.assertFalse(phys_brs.called)
            self.assertFalse(reset_tun_br.called)
            self.assertFalse(tun_br.called)
            self.assertFalse(logger_info.called)
            self.assertTrue(exception_log.called)
            self.assertFalse(self.agent.refresh_firewall_required)
            self.assertEqual(0, len(self.agent.devices_to_filter))

    def _get_fake_port(self, port_id):
        return {'id': port_id,
                'port_id': port_id,
                'segmentation_id': 1232,
                'lvid': 1,
                'network_id': 'fake_network',
                'device': port_id,
                'admin_state_up': True}

    def test_update_port_dict(self):
        fakeport = self._get_fake_port('fakeId')
        self.agent.ports_dict = {}
        self.agent.network_port_count = {}
        self.agent.tenant_network_type = p_const.TYPE_VLAN
        with mock.patch.object(self.agent.sg_agent, 'add_devices_to_filter'
                               ) as mock_add_devices:
            status = self.agent._update_port_dict(fakeport)
            self.assertIn('fakeId', self.agent.ports_dict)
            self.assertTrue(status)
            self.assertEqual(1, self.agent.network_port_count['fake_network'])
            mock_add_devices.assert_called_with([fakeport])

    def test_update_port_dict_existing_network(self):
        fakeport = self._get_fake_port('fakeId')
        self.agent.ports_dict = {}
        self.agent.network_port_count = {'fake_network': 6}
        self.agent.tenant_network_type = p_const.TYPE_VLAN
        with mock.patch.object(self.agent.sg_agent, 'add_devices_to_filter'
                               ) as mock_add_devices:
            status = self.agent._update_port_dict(fakeport)
            self.assertIn('fakeId', self.agent.ports_dict)
            self.assertTrue(status)
            self.assertEqual(7, self.agent.network_port_count['fake_network'])
            mock_add_devices.assert_called_with([fakeport])

    def test_update_firewall(self):
        fakeport_1 = self._get_fake_port(FAKE_PORT_1)
        fakeport_2 = self._get_fake_port(FAKE_PORT_2)
        self.agent.devices_to_filter = set([FAKE_PORT_1,
                                            FAKE_PORT_2])
        self.agent.ports_dict = {FAKE_PORT_1: fakeport_1}
        self.agent.refresh_firewall_required = True
        self.agent.tenant_network_type = p_const.TYPE_VLAN
        self.agent.vcenter_id = FAKE_VCENTER
        self.agent.cluster_id = FAKE_CLUSTER_1
        with contextlib.nested(
            mock.patch.object(self.agent.ovsvapp_rpc,
                              'get_ports_details_list',
                              return_value=[fakeport_2]),
            mock.patch.object(self.agent.sg_agent, 'refresh_firewall'),
        ) as (mock_get_ports_details_list, mock_refresh_firewall):
            self.agent._update_firewall()
            self.assertFalse(self.agent.refresh_firewall_required)
            self.assertFalse(self.agent.devices_to_filter)
            self.assertIn(FAKE_PORT_2, self.agent.ports_dict)
            mock_get_ports_details_list.assert_called_with(
                self.agent.context,
                set([FAKE_PORT_2]),
                self.agent.agent_id,
                self.agent.vcenter_id,
                self.agent.cluster_id)
            mock_refresh_firewall.assert_called_with(set([FAKE_PORT_1,
                                                          FAKE_PORT_2]))

    def test_update_firewall_get_ports_exception(self):
        fakeport_1 = self._get_fake_port(FAKE_PORT_1)
        self.agent.devices_to_filter = set([FAKE_PORT_1,
                                            FAKE_PORT_2])
        self.agent.ports_dict = {FAKE_PORT_1: fakeport_1}
        self.agent.refresh_firewall_required = True
        self.agent.vcenter_id = FAKE_VCENTER
        self.agent.cluster_id = FAKE_CLUSTER_1
        with contextlib.nested(
            mock.patch.object(self.agent.ovsvapp_rpc,
                              'get_ports_details_list',
                              side_effect=Exception()),
            mock.patch.object(self.agent.sg_agent, 'refresh_firewall'),
        ) as (mock_get_ports_details_list, mock_refresh_firewall):
            self.agent._update_firewall()
            self.assertTrue(self.agent.refresh_firewall_required)
            self.assertEqual(set([FAKE_PORT_2]), self.agent.devices_to_filter)
            self.assertNotIn(FAKE_PORT_2, self.agent.ports_dict)
            mock_get_ports_details_list.assert_called_with(
                self.agent.context,
                set([FAKE_PORT_2]),
                self.agent.agent_id,
                self.agent.vcenter_id,
                self.agent.cluster_id)
            mock_refresh_firewall.assert_called_with(set([FAKE_PORT_1]))

    def test_check_for_updates_no_updates(self):
        self.agent.refresh_firewall_required = False
        self.agent.ports_to_bind = None
        with contextlib.nested(
            mock.patch.object(self.agent, 'check_ovs_status',
                              return_value=4),
            mock.patch.object(self.agent, '_update_firewall'),
            mock.patch.object(self.agent.sg_agent,
                              'firewall_refresh_needed',
                              return_value=False),
            mock.patch.object(self.agent.sg_agent, 'refresh_port_filters'),
            mock.patch.object(self.agent, '_update_port_bindings')
        ) as (mock_check_ovs, mock_update_firewall, mock_firewall_refresh,
              mock_refresh_port_filters, mock_update_port_bindings):
            self.agent._check_for_updates()
            self.assertTrue(mock_check_ovs.called)
            self.assertFalse(mock_update_firewall.called)
            self.assertTrue(mock_firewall_refresh.called)
            self.assertFalse(mock_refresh_port_filters.called)
            self.assertFalse(mock_update_port_bindings.called)

    def test_check_for_updates_ovs_restarted(self):
        self.agent.refresh_firewall_required = False
        self.agent.ports_to_bind = None
        with contextlib.nested(
            mock.patch.object(self.agent, 'check_ovs_status',
                              return_value=0),
            mock.patch.object(self.agent, 'mitigate_ovs_restart'),
            mock.patch.object(self.agent, '_update_firewall'),
            mock.patch.object(self.agent.sg_agent,
                              'firewall_refresh_needed',
                              return_value=False),
            mock.patch.object(self.agent, '_update_port_bindings')
        ) as (mock_check_ovs, mock_mitigate, mock_update_firewall,
              mock_firewall_refresh, mock_update_port_bindings):
            self.agent._check_for_updates()
            self.assertTrue(mock_check_ovs.called)
            self.assertTrue(mock_mitigate.called)
            self.assertFalse(mock_update_firewall.called)
            self.assertTrue(mock_firewall_refresh.called)
            self.assertFalse(mock_update_port_bindings.called)

    def test_check_for_updates_devices_to_filter(self):
        self.agent.refresh_firewall_required = True
        self.agent.ports_to_bind = None
        with contextlib.nested(
            mock.patch.object(self.agent, 'check_ovs_status',
                              return_value=4),
            mock.patch.object(self.agent, 'mitigate_ovs_restart'),
            mock.patch.object(self.agent, '_update_firewall'),
            mock.patch.object(self.agent.sg_agent,
                              'firewall_refresh_needed',
                              return_value=False),
            mock.patch.object(self.agent, '_update_port_bindings')
        ) as (mock_check_ovs, mock_mitigate, mock_update_firewall,
              mock_firewall_refresh, mock_update_port_bindings):
            self.agent._check_for_updates()
            self.assertTrue(mock_check_ovs.called)
            self.assertFalse(mock_mitigate.called)
            self.assertTrue(mock_update_firewall.called)
            self.assertTrue(mock_firewall_refresh.called)
            self.assertFalse(mock_update_port_bindings.called)

    def test_check_for_updates_firewall_refresh(self):
        self.agent.refresh_firewall_required = False
        self.agent.ports_to_bind = None
        with contextlib.nested(
            mock.patch.object(self.agent, 'check_ovs_status',
                              return_value=4),
            mock.patch.object(self.agent, '_update_firewall'),
            mock.patch.object(self.agent.sg_agent,
                              'firewall_refresh_needed',
                              return_value=True),
            mock.patch.object(self.agent.sg_agent, 'refresh_port_filters'),
            mock.patch.object(self.agent, '_update_port_bindings')
        ) as (mock_check_ovs, mock_update_firewall, mock_firewall_refresh,
              mock_refresh_port_filters, mock_update_port_bindings):
            self.agent._check_for_updates()
            self.assertTrue(mock_check_ovs.called)
            self.assertFalse(mock_update_firewall.called)
            self.assertTrue(mock_firewall_refresh.called)
            self.assertTrue(mock_refresh_port_filters.called)
            self.assertFalse(mock_update_port_bindings.called)

    def test_check_for_updates_port_bindings(self):
        self.agent.refresh_firewall_required = False
        self.agent.ports_to_bind.add("fake_port")
        with contextlib.nested(
            mock.patch.object(self.agent, 'check_ovs_status',
                              return_value=4),
            mock.patch.object(self.agent, '_update_firewall'),
            mock.patch.object(self.agent.sg_agent,
                              'firewall_refresh_needed',
                              return_value=False),
            mock.patch.object(self.agent, '_update_port_bindings')
        ) as (mock_check_ovs, mock_update_firewall, mock_firewall_refresh,
              mock_update_port_bindings):
            self.agent._check_for_updates()
            self.assertTrue(mock_check_ovs.called)
            self.assertFalse(mock_update_firewall.called)
            self.assertTrue(mock_firewall_refresh.called)
            self.assertTrue(mock_update_port_bindings.called)

    def test_report_state(self):
        with mock.patch.object(self.agent.state_rpc,
                               "report_state") as report_st:
            self.agent._report_state()
            report_st.assert_called_with(self.agent.context,
                                         self.agent.agent_state,
                                         True)
            self.assertNotIn("start_flag", self.agent.agent_state)
            self.assertFalse(self.agent.use_call)
            self.assertEqual(cfg.CONF.host,
                             self.agent.agent_state["host"])

    def test_report_state_fail(self):
        with contextlib.nested(
            mock.patch.object(self.agent.state_rpc,
                              "report_state",
                              side_effect=Exception()),
            mock.patch.object(self.LOG, 'exception'),
        ) as (report_st, log_exception):
            self.agent._report_state()
            report_st.assert_called_with(self.agent.context,
                                         self.agent.agent_state,
                                         True)
            self.assertTrue(log_exception.called)

    def test_process_event_ignore_event(self):
        vm = VM(FAKE_VM, [])
        event = SampleEvent(VNIC_ADDED, FAKE_HOST_1,
                            FAKE_CLUSTER_MOID, vm)
        with contextlib.nested(
            mock.patch.object(self.agent,
                              "_notify_device_added"),
            mock.patch.object(self.agent,
                              "_notify_device_updated"),
            mock.patch.object(self.agent,
                              "_notify_device_deleted"),
            mock.patch.object(self.LOG, 'debug')
        ) as (add_vm, update_vm, del_vm, log_debug):
            self.agent.process_event(event)
            self.assertFalse(add_vm.called)
            self.assertFalse(update_vm.called)
            self.assertFalse(del_vm.called)
            self.assertTrue(log_debug.called)

    def test_process_event_exception(self):
        vm = VM(FAKE_VM, [])
        event = SampleEvent(ovsvapp_const.VM_CREATED,
                            FAKE_HOST_1, FAKE_CLUSTER_MOID, vm)
        with contextlib.nested(
            mock.patch.object(self.agent,
                              "_notify_device_added",
                              side_effect=Exception()),
            mock.patch.object(self.LOG, 'exception'),
            mock.patch.object(self.LOG, 'error'),
        ) as (add_vm, log_exception, log_error):
            self.agent.process_event(event)
            self.assertTrue(add_vm.called)
            self.assertTrue(log_error.called)
            self.assertTrue(log_exception.called)

    def test_process_event_vm_create_nonics_non_host_non_cluster(self):
        self.agent.esx_hostname = FAKE_HOST_2
        vm = VM(FAKE_VM, [])
        event = SampleEvent(ovsvapp_const.VM_CREATED,
                            FAKE_HOST_1, FAKE_CLUSTER_MOID, vm)
        self.agent.state = ovsvapp_const.AGENT_RUNNING
        with mock.patch.object(self.agent,
                               "_notify_device_added") as device_added:
            self.agent.process_event(event)
            self.assertTrue(device_added.called)

    def test_process_event_vm_create_nonics_non_host(self):
        self.agent.esx_hostname = FAKE_HOST_2
        vm = VM(FAKE_VM, [])
        event = SampleEvent(ovsvapp_const.VM_CREATED,
                            FAKE_HOST_1, FAKE_CLUSTER_MOID, vm)
        self.agent.state = ovsvapp_const.AGENT_RUNNING
        with mock.patch.object(self.agent,
                               "_notify_device_added") as device_added:
            self.agent.process_event(event)
            self.assertTrue(device_added.called)
            self.assertEqual(FAKE_CLUSTER_MOID, self.agent.cluster_moid)

    def test_process_event_vm_create_nics_non_host(self):
        self.agent.esx_hostname = FAKE_HOST_2
        vm_port1 = SamplePort(FAKE_PORT_1)
        vm_port2 = SamplePort(FAKE_PORT_2)
        vm = VM(FAKE_VM, ([vm_port1, vm_port2]))
        event = SampleEvent(ovsvapp_const.VM_CREATED,
                            FAKE_HOST_1, FAKE_CLUSTER_MOID, vm)
        self.agent.state = ovsvapp_const.AGENT_RUNNING
        self.agent.process_event(event)
        for vnic in vm.vnics:
            self.assertIn(vnic.port_uuid, self.agent.devices_to_filter)
            self.assertIn(vnic.port_uuid, self.agent.cluster_other_ports)
            self.assertNotIn(vnic.port_uuid, self.agent.cluster_host_ports)

    def test_process_event_vm_create_nics_host(self):
        self.agent.esx_hostname = FAKE_HOST_1
        vm_port1 = SamplePort(FAKE_PORT_1)
        vm_port2 = SamplePort(FAKE_PORT_2)
        vm = VM(FAKE_VM, ([vm_port1, vm_port2]))
        event = SampleEvent(ovsvapp_const.VM_CREATED,
                            FAKE_HOST_1, FAKE_CLUSTER_MOID, vm)
        self.agent.state = ovsvapp_const.AGENT_RUNNING
        self.agent.process_event(event)
        for vnic in vm.vnics:
            self.assertIn(vnic.port_uuid, self.agent.devices_to_filter)
            self.assertIn(vnic.port_uuid, self.agent.cluster_host_ports)
            self.assertNotIn(vnic.port_uuid, self.agent.cluster_other_ports)

    def test_process_event_vm_updated_nonhost(self):
        self.agent.esx_hostname = FAKE_HOST_2
        vm_port1 = SamplePort(FAKE_PORT_1)
        vm = VM(FAKE_VM, [vm_port1])
        event = SampleEvent(ovsvapp_const.VM_UPDATED,
                            FAKE_HOST_1, FAKE_CLUSTER_MOID, vm, True)
        self.agent.state = ovsvapp_const.AGENT_RUNNING
        self.agent.tenant_network_type = p_const.TYPE_VLAN
        self.agent.process_event(event)
        self.assertIn(FAKE_PORT_1, self.agent.cluster_other_ports)

    def test_process_event_vm_delete_hosted_vm_vlan(self):
        self.agent.esx_hostname = FAKE_HOST_1
        self.agent.cluster_host_ports.add(FAKE_PORT_1)
        self.agent.tenant_network_type = p_const.TYPE_VLAN
        self.agent.network_port_count['net_uuid'] = 1
        port = self._build_port()
        self.agent.ports_dict[port['id']] = self.agent._build_port_info(
            port)
        del_port = self.agent.ports_dict[port['id']]
        vm_port = SamplePortUIDMac(FAKE_PORT_1, MAC_ADDRESS)
        vm = VM(FAKE_VM, ([vm_port]))
        event = SampleEvent(ovsvapp_const.VM_DELETED,
                            FAKE_HOST_1, FAKE_CLUSTER_MOID, vm)
        self.agent.state = ovsvapp_const.AGENT_RUNNING
        self.agent.net_mgr = fake_manager.MockNetworkManager("callback")
        self.agent.net_mgr.initialize_driver()
        with contextlib.nested(
            mock.patch.object(self.agent.net_mgr.get_driver(),
                              "post_delete_vm",
                              return_value=True),
            mock.patch.object(self.agent.net_mgr.get_driver(),
                              "delete_network"),
        ) as (post_del_vm, del_net):
            self.agent.process_event(event)
            for vnic in vm.vnics:
                self.assertNotIn(vnic.port_uuid, self.agent.cluster_host_ports)
            self.assertTrue(post_del_vm.called)
            self.assertTrue(del_net.called)
            self.assertNotIn(del_port.network_id,
                             self.agent.network_port_count.keys())

    def test_process_event_vm_delete_hosted_vm_vxlan(self):
        self.agent.esx_hostname = FAKE_HOST_1
        self.agent.cluster_host_ports.add(FAKE_PORT_1)
        self.agent.tenant_network_type = p_const.TYPE_VXLAN
        self.agent.network_port_count['net_uuid'] = 1
        port = self._build_port()
        self.agent.ports_dict[port['id']] = self.agent._build_port_info(
            port)
        vm_port = SamplePortUIDMac(FAKE_PORT_1, MAC_ADDRESS)
        vm = VM(FAKE_VM, ([vm_port]))
        event = SampleEvent(ovsvapp_const.VM_DELETED,
                            FAKE_HOST_1, FAKE_CLUSTER_MOID, vm)
        self.agent.state = ovsvapp_const.AGENT_RUNNING
        self.agent.net_mgr = fake_manager.MockNetworkManager("callback")
        self.agent.net_mgr.initialize_driver()
        with mock.patch.object(self.agent.net_mgr.get_driver(),
                               "post_delete_vm",
                               return_value=True) as (post_del_vm):
            self.agent.process_event(event)
            for vnic in vm.vnics:
                self.assertNotIn(vnic.port_uuid, self.agent.cluster_host_ports)
            self.assertTrue(post_del_vm.called)

    def test_process_event_vm_delete_non_hosted_vm(self):
        self.agent.esx_hostname = FAKE_HOST_2
        self.agent.cluster_other_ports.add(FAKE_PORT_1)
        self.agent.tenant_network_type = p_const.TYPE_VLAN
        self.agent.network_port_count['net_uuid'] = 1
        port = self._build_port()
        self.agent.ports_dict[port['id']] = self.agent._build_port_info(
            port)
        del_port = self.agent.ports_dict[port['id']]
        vm_port = SamplePortUIDMac(FAKE_PORT_1, MAC_ADDRESS)
        vm = VM(FAKE_VM, ([vm_port]))
        event = SampleEvent(ovsvapp_const.VM_DELETED,
                            FAKE_HOST_1, FAKE_CLUSTER_MOID, vm)
        self.agent.net_mgr = fake_manager.MockNetworkManager("callback")
        self.agent.net_mgr.initialize_driver()
        self.agent.state = ovsvapp_const.AGENT_RUNNING
        with contextlib.nested(
            mock.patch.object(self.agent.net_mgr.get_driver(),
                              "post_delete_vm",
                              return_value=True),
            mock.patch.object(self.agent.net_mgr.get_driver(),
                              "delete_network"),
        ) as (post_del_vm, del_net):
            self.agent.process_event(event)
            for vnic in vm.vnics:
                self.assertNotIn(vnic.port_uuid,
                                 self.agent.cluster_other_ports)
            self.assertTrue(post_del_vm.called)
            self.assertFalse(del_net.called)
            self.assertNotIn(del_port.network_id,
                             self.agent.network_port_count.keys())

    def test_notify_device_added_with_hosted_vm(self):
        vm = VM(FAKE_VM, [])
        host = FAKE_HOST_1
        self.agent.esx_hostname = host
        self.agent.state = ovsvapp_const.AGENT_RUNNING
        with contextlib.nested(
            mock.patch.object(self.agent.ovsvapp_rpc,
                              "get_ports_for_device",
                              return_value=True),
            mock.patch.object(self.LOG, 'exception'),
            mock.patch.object(time, "sleep"),
        ) as (get_ports, log_exception, time_sleep):
            self.agent._notify_device_added(vm, host)
            self.assertTrue(get_ports.called)
            self.assertFalse(time_sleep.called)
            self.assertFalse(log_exception.called)

    def test_notify_device_added_rpc_exception(self):
        vm = VM(FAKE_VM, [])
        host = FAKE_HOST_1
        self.agent.esx_hostname = host
        self.agent.state = ovsvapp_const.AGENT_RUNNING
        with contextlib.nested(
            mock.patch.object(self.agent.ovsvapp_rpc,
                              "get_ports_for_device",
                              side_effect=Exception()),
            mock.patch.object(self.LOG, 'exception'),
            mock.patch.object(time, "sleep"),
        ) as (get_ports, log_exception, time_sleep):
            self.assertRaises(
                error.OVSvAppNeutronAgentError,
                self.agent._notify_device_added, vm, host)
            self.assertTrue(log_exception.called)
            self.assertTrue(get_ports.called)
            self.assertFalse(time_sleep.called)

    def test_notify_device_added_with_retry(self):
        vm = VM(FAKE_VM, [])
        host = FAKE_HOST_1
        self.agent.esx_hostname = host
        self.agent.state = ovsvapp_const.AGENT_RUNNING
        with contextlib.nested(
            mock.patch.object(self.agent.ovsvapp_rpc,
                              "get_ports_for_device",
                              return_value=False),
            mock.patch.object(self.LOG, 'exception'),
            mock.patch.object(time, "sleep"),
        ) as (get_ports, log_exception, time_sleep):
            self.agent._notify_device_added(vm, host)
            self.assertTrue(get_ports.called)
            self.assertTrue(time_sleep.called)
            self.assertFalse(log_exception.called)

    def test_notify_device_updated_host(self):
        host = FAKE_HOST_1
        self.agent.esx_hostname = host
        vm_port1 = SamplePort(FAKE_PORT_1)
        vm = VM(FAKE_VM, [vm_port1])
        self.agent.state = ovsvapp_const.AGENT_RUNNING
        self.agent.tenant_network_type = p_const.TYPE_VXLAN
        with contextlib.nested(
            mock.patch.object(self.agent.ovsvapp_rpc,
                              "update_port_binding"),
            mock.patch.object(self.agent.plugin_rpc,
                              "get_device_details"),
            mock.patch.object(self.agent.plugin_rpc,
                              "update_device_up"),
            mock.patch.object(self.LOG, 'exception'),
        ) as (update_port_binding, get_device_details,
              update_device_up, log_exception):
            self.agent._notify_device_updated(vm, host, True)
            self.assertTrue(update_port_binding.called)
            self.assertTrue(get_device_details.called)
            self.assertTrue(update_device_up.called)
            self.assertIn(FAKE_PORT_1, self.agent.cluster_host_ports)
            self.assertFalse(log_exception.called)

    def test_notify_device_updated_rpc_exception(self):
        host = FAKE_HOST_1
        self.agent.esx_hostname = host
        vm_port1 = SamplePort(FAKE_PORT_1)
        vm = VM(FAKE_VM, [vm_port1])
        self.agent.state = ovsvapp_const.AGENT_RUNNING
        self.agent.tenant_network_type = p_const.TYPE_VXLAN
        with contextlib.nested(
            mock.patch.object(self.agent.ovsvapp_rpc,
                              "update_port_binding",
                              side_effect=Exception()),
            mock.patch.object(self.agent.plugin_rpc,
                              "get_device_details"),
            mock.patch.object(self.agent.plugin_rpc,
                              "update_device_up"),
            mock.patch.object(self.LOG, 'exception'),
        ) as (update_port_binding, get_device_details,
              update_device_up, log_exception):
            self.assertRaises(
                error.OVSvAppNeutronAgentError,
                self.agent._notify_device_updated, vm, host, True)
            self.assertTrue(update_port_binding.called)
            self.assertFalse(get_device_details.called)
            self.assertFalse(update_device_up.called)
            self.assertIn(FAKE_PORT_1, self.agent.cluster_host_ports)
            self.assertTrue(log_exception.called)

    def test_map_port_to_common_model_vlan(self):
        expected_port = self._build_port()
        self.agent.tenant_network_type = p_const.TYPE_VLAN
        network, port = self.agent._map_port_to_common_model(expected_port)
        self.assertEqual(expected_port['network_id'], network.name)
        self.assertEqual(expected_port['id'], port.uuid)

    def test_map_port_to_common_model_vxlan(self):
        expected_port = self._build_port()
        self.agent.cluster_moid = FAKE_CLUSTER_MOID
        self.agent.tenant_network_type = p_const.TYPE_VXLAN
        network, port = self.agent._map_port_to_common_model(expected_port, 1)
        expected_name = expected_port['network_id'] + "-" + FAKE_CLUSTER_MOID
        self.assertEqual(expected_name, network.name)
        self.assertEqual(expected_port['id'], port.uuid)

    def test_device_create_cluster_mismatch(self):
        self.agent.vcenter_id = FAKE_VCENTER
        self.agent.cluster_id = FAKE_CLUSTER_2
        with contextlib.nested(
            mock.patch.object(self.agent,
                              '_process_create_portgroup_vlan',
                              return_value=True),
            mock.patch.object(self.LOG, 'debug')
        ) as (create_pg_vlan, logger_debug):
            self.agent.device_create(FAKE_CONTEXT,
                                     device=DEVICE)
            self.assertTrue(logger_debug.called)
            self.assertFalse(create_pg_vlan.called)

    def test_device_create_non_hosted_vm(self):
        ports = [self._build_port()]
        self.agent.vcenter_id = FAKE_VCENTER
        self.agent.cluster_id = FAKE_CLUSTER_1
        self.agent.esx_hostname = FAKE_HOST_2
        self.agent.tenant_network_type = p_const.TYPE_VLAN
        with contextlib.nested(
            mock.patch.object(self.agent.plugin_rpc, 'update_device_up',
                              return_value=True),
            mock.patch.object(self.agent.sg_agent, 'add_devices_to_filter'),
            mock.patch.object(self.agent.sg_agent, 'ovsvapp_sg_update'),
            mock.patch.object(self.LOG, 'debug')
        ) as (update_device_up, mock_add_devices_fn, mock_sg_update_fn,
              logger_debug):
            self.agent.device_create(FAKE_CONTEXT,
                                     device=DEVICE,
                                     ports=ports,
                                     sg_rules=FAKE_SG_RULES)
            self.assertTrue(logger_debug.called)
            mock_add_devices_fn.assert_called_with(ports)
            self.assertIn(FAKE_PORT_1, self.agent.cluster_other_ports)
            self.assertNotIn(FAKE_PORT_1, self.agent.cluster_host_ports)
            self.assertFalse(mock_sg_update_fn.called)
            self.assertFalse(update_device_up.called)

    def test_device_create_hosted_vm_vlan(self):
        ports = [self._build_port()]
        self.agent.vcenter_id = FAKE_VCENTER
        self.agent.cluster_id = FAKE_CLUSTER_1
        self.agent.esx_hostname = FAKE_HOST_1
        self.agent.tenant_network_type = p_const.TYPE_VLAN
        self.agent.net_mgr = fake_manager.MockNetworkManager("callback")
        self.agent.net_mgr.initialize_driver()
        with contextlib.nested(
            mock.patch.object(self.agent.plugin_rpc, 'update_device_up',
                              return_value=True),
            mock.patch.object(self.agent.sg_agent, 'add_devices_to_filter'),
            mock.patch.object(self.agent.sg_agent, 'ovsvapp_sg_update'),
            mock.patch.object(self.LOG, 'debug')
        ) as (update_device_up, mock_add_devices_fn, mock_sg_update_fn,
              logger_debug):
            self.agent.device_create(FAKE_CONTEXT,
                                     device=DEVICE,
                                     ports=ports,
                                     sg_rules=FAKE_SG_RULES)
            self.assertTrue(logger_debug.called)
            self.assertNotIn(FAKE_PORT_1, self.agent.cluster_other_ports)
            self.assertIn(FAKE_PORT_1, self.agent.cluster_host_ports)
            mock_add_devices_fn.assert_called_with(ports)
            mock_sg_update_fn.assert_called_with(
                FAKE_SG_RULES.get(FAKE_DEVICE_ID))
            self.assertTrue(update_device_up.called)

    def test_device_create_hosted_vm_vxlan(self):
        ports = [self._build_port()]
        self.agent.vcenter_id = FAKE_VCENTER
        self.agent.cluster_id = FAKE_CLUSTER_1
        self.agent.cluster_moid = FAKE_CLUSTER_MOID
        self.agent.esx_hostname = FAKE_HOST_1
        self.agent.tenant_network_type = p_const.TYPE_VXLAN
        self.agent.local_vlan_map = {}
        self.agent.net_mgr = fake_manager.MockNetworkManager("callback")
        self.agent.net_mgr.initialize_driver()
        with contextlib.nested(
            mock.patch.object(self.agent, '_populate_tunnel_flows_for_port'),
            mock.patch.object(self.agent.plugin_rpc, 'update_device_up',
                              return_value=True),
            mock.patch.object(self.agent.sg_agent, 'add_devices_to_filter'),
            mock.patch.object(self.agent.sg_agent, 'ovsvapp_sg_update'),
            mock.patch.object(self.LOG, 'debug')
        ) as (populate_tun_flows, update_device_up, mock_add_devices_fn,
              mock_sg_update_fn, logger_debug):
            self.agent.device_create(FAKE_CONTEXT,
                                     device=DEVICE,
                                     ports=ports,
                                     sg_rules=FAKE_SG_RULES)
            self.assertTrue(populate_tun_flows.called)
            self.assertTrue(logger_debug.called)
            self.assertNotIn(FAKE_PORT_1, self.agent.cluster_other_ports)
            self.assertIn(FAKE_PORT_1, self.agent.cluster_host_ports)
            mock_add_devices_fn.assert_called_with(ports)
            mock_sg_update_fn.assert_called_with(
                FAKE_SG_RULES.get(FAKE_DEVICE_ID))
            self.assertTrue(update_device_up.called)

    def test_device_create_hosted_vm_create_port_exception(self):
        ports = [self._build_port()]
        self.agent.vcenter_id = FAKE_VCENTER
        self.agent.cluster_id = FAKE_CLUSTER_1
        self.agent.esx_hostname = FAKE_HOST_1
        self.agent.tenant_network_type = p_const.TYPE_VLAN
        self.agent.net_mgr = fake_manager.MockNetworkManager("callback")
        self.agent.net_mgr.initialize_driver()
        self.agent.net_mgr.get_driver().create_port = mock.Mock(
            side_effect=Exception())
        with contextlib.nested(
            mock.patch.object(self.agent.plugin_rpc, 'update_device_up'),
            mock.patch.object(self.agent.sg_agent, 'add_devices_to_filter'),
            mock.patch.object(self.agent.sg_agent, 'ovsvapp_sg_update'),
            mock.patch.object(self.LOG, 'debug'),
            mock.patch.object(self.LOG, 'exception'),
        ) as (update_device, mock_add_devices_fn, mock_sg_update_fn,
              logger_debug, log_excep):
            self.assertRaises(
                error.OVSvAppNeutronAgentError,
                self.agent.device_create,
                FAKE_CONTEXT, device=DEVICE,
                ports=ports, sg_rules=FAKE_SG_RULES)
            self.assertTrue(logger_debug.called)
            self.assertNotIn(FAKE_PORT_1, self.agent.cluster_other_ports)
            self.assertIn(FAKE_PORT_1, self.agent.cluster_host_ports)
            self.assertFalse(update_device.called)
            self.assertFalse(mock_sg_update_fn.called)
            self.assertTrue(log_excep.called)

    def test_device_create_hosted_vm_update_device_exception(self):
        ports = [self._build_port()]
        self.agent.vcenter_id = FAKE_VCENTER
        self.agent.cluster_id = FAKE_CLUSTER_1
        self.agent.esx_hostname = FAKE_HOST_1
        self.agent.tenant_network_type = p_const.TYPE_VLAN
        self.agent.net_mgr = fake_manager.MockNetworkManager("callback")
        self.agent.net_mgr.initialize_driver()
        with contextlib.nested(
            mock.patch.object(self.agent.sg_agent, 'add_devices_to_filter'),
            mock.patch.object(self.agent.sg_agent, 'ovsvapp_sg_update'),
            mock.patch.object(self.agent.plugin_rpc, 'update_device_up',
                              side_effect=Exception()),
            mock.patch.object(self.LOG, 'debug'),
            mock.patch.object(self.LOG, 'exception'),
        ) as (mock_add_devices_fn, mock_sg_update_fn, update_device,
              logger_debug, log_excep):
            self.assertRaises(
                error.OVSvAppNeutronAgentError,
                self.agent.device_create,
                FAKE_CONTEXT, device=DEVICE,
                ports=ports, sg_rules=FAKE_SG_RULES)
            self.assertTrue(logger_debug.called)
            self.assertNotIn(FAKE_PORT_1, self.agent.cluster_other_ports)
            self.assertIn(FAKE_PORT_1, self.agent.cluster_host_ports)
            mock_add_devices_fn.assert_called_with(ports)
            self.assertFalse(mock_sg_update_fn.called)
            self.assertTrue(update_device.called)
            self.assertTrue(log_excep.called)

    def test_port_update_admin_state_up(self):
        port = self._build_port()
        self.agent.ports_dict[port['id']] = self.agent._build_port_info(
            port)
        self.agent.cluster_host_ports = set([port['id']])
        self.agent.tenant_network_type = p_const.TYPE_VLAN
        self.agent.net_mgr = fake_manager.MockNetworkManager("callback")
        self.agent.net_mgr.initialize_driver()
        port['admin_state_up'] = True
        neutron_port = {'port': port,
                        'segmentation_id': port['segmentation_id']}
        with contextlib.nested(
            mock.patch.object(self.agent.plugin_rpc,
                              "update_device_up"),
            mock.patch.object(self.agent.plugin_rpc,
                              "update_device_down"),
            mock.patch.object(self.LOG, 'exception'),
            mock.patch.object(self.LOG, 'debug')
        ) as (device_up, device_down,
              log_exception, log_debug):
            self.agent.port_update(FAKE_CONTEXT, **neutron_port)
            self.assertEqual(neutron_port['port']['admin_state_up'],
                             self.agent.ports_dict[port['id']].
                             admin_state_up)
            self.assertTrue(device_up.called)
            self.assertFalse(device_down.called)
            self.assertFalse(log_exception.called)

    def test_port_update_rpc_exception(self):
        port = self._build_port()
        self.agent.ports_dict[port['id']] = self.agent._build_port_info(
            port)
        self.agent.cluster_host_ports = set([port['id']])
        self.agent.tenant_network_type = p_const.TYPE_VLAN
        self.agent.net_mgr = fake_manager.MockNetworkManager("callback")
        self.agent.net_mgr.initialize_driver()
        port['admin_state_up'] = True
        neutron_port = {'port': port,
                        'segmentation_id': port['segmentation_id']}
        with contextlib.nested(
            mock.patch.object(self.agent.plugin_rpc,
                              "update_device_up",
                              side_effect=Exception()),
            mock.patch.object(self.agent.plugin_rpc,
                              "update_device_down"),
            mock.patch.object(self.LOG, 'exception'),
            mock.patch.object(self.LOG, 'debug')
        ) as (device_up, device_down,
              log_exception, log_debug):
            self.assertRaises(
                error.OVSvAppNeutronAgentError,
                self.agent.port_update, FAKE_CONTEXT, **neutron_port)
            self.assertEqual(neutron_port['port']['admin_state_up'],
                             self.agent.ports_dict[port['id']].
                             admin_state_up)
            self.assertTrue(device_up.called)
            self.assertFalse(device_down.called)
            self.assertTrue(log_exception.called)



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib

import mock
from oslo_config import cfg

from networking_vsphere.agent import agent
from networking_vsphere.common import constants
from networking_vsphere.common import utils
from networking_vsphere.tests import base
from networking_vsphere.tests.unit.drivers import fake_manager

CONF = cfg.CONF


class TestAgent(base.TestCase):

    def setUp(self):
        super(TestAgent, self).setUp()
        self.agent = agent.Agent()
        self.LOG = agent.LOG

    def test_initialize_managers(self):
        fake_net_mgr = fake_manager.MockNetworkManager(self.agent)
        self.assertIsNone(fake_net_mgr.get_driver())
        with mock.patch.object(utils, "load_object",
                               return_value=fake_net_mgr):
            self.agent._initialize_managers()
            self.assertTrue(fake_net_mgr.get_driver())
            self.assertTrue(self.agent.state == constants.AGENT_INITIALIZED)

    def test_set_node_state_already_up(self):
        self.agent.node_up = True
        with mock.patch.object(self.LOG, 'info') as log_info:
            self.agent.set_node_state(True)
            self.assertTrue(log_info.called)

    def test_set_node_state_up(self):
        fake_net_mgr = fake_manager.MockNetworkManager(self.agent)
        self.assertIsNone(fake_net_mgr.get_driver())
        with contextlib.nested(
            mock.patch.object(utils, "load_object",
                              return_value=fake_net_mgr),
            mock.patch.object(self.LOG, 'info')
        ) as (load_obj, log_info):
            self.agent.set_node_state(True)
            self.assertTrue(log_info.called)
            self.assertTrue(fake_net_mgr.get_driver())
            self.assertTrue("start" in fake_net_mgr.methods)
            self.assertTrue(self.agent.state == constants.AGENT_RUNNING)

    def test_set_node_state_down(self):
        self.agent.node_up = True
        fake_net_mgr = fake_manager.MockNetworkManager(self.agent)
        self.agent.net_mgr = fake_net_mgr
        self.agent.set_node_state(False)
        self.assertTrue("stop" in fake_net_mgr.methods)
        self.assertTrue(self.agent.state == constants.AGENT_STOPPED)



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib

import mock
from oslo_vmware import api
from oslo_vmware import vim

from networking_vsphere.tests import base
from networking_vsphere.utils import vim_session


class TestVmwareApiSession(base.TestCase):

    def setUp(self):
        super(TestVmwareApiSession, self).setUp()
        self.host_ip = "192.168.1.3"
        self.host_username = "user"
        self.host_password = "password"
        self.api_retry_count = 2
        self.wsdl_url = "fake_url"
        self.vm_session = vim_session.VMWareAPISession(self.host_ip,
                                                       self.host_username,
                                                       self.host_password,
                                                       self.api_retry_count,
                                                       self.wsdl_url,
                                                       create_session=False)

    def test_call_method(self):
        with contextlib.nested(
            mock.patch.object(api.VMwareAPISession,
                              "invoke_api"),
            mock.patch.object(self.vm_session,
                              "_is_vim_object",
                              return_value=True)
            ) as (invoke_ob, is_vim_ob):
            self.vm_session._call_method("fake_module",
                                         "get_objects",
                                         "HostSystem", ['name'])
            self.assertTrue(invoke_ob.called)

    def test_call_method_with_vim_object_false(self):
        vim.Vim = mock.Mock()
        with contextlib.nested(
            mock.patch.object(api.VMwareAPISession,
                              "invoke_api"),
            mock.patch.object(api.VMwareAPISession,
                              "vim",
                              return_value=vim.Vim),
            mock.patch.object(self.vm_session,
                              "_is_vim_object",
                              return_value=False)
            ) as (invoke_ob, vim_prop, is_vim_ob):
            self.vm_session._call_method("fake_module",
                                         "get_objects",
                                         "HostSystem", ['name'])
            self.assertTrue(invoke_ob.called)

    def test_get_vim(self):
        vim.Vim = mock.Mock(return_value="fake_vim")
        with contextlib.nested(
            mock.patch.object(api.VMwareAPISession,
                              "vim",
                              return_value=vim.Vim)):
            new_vim = self.vm_session._get_vim()
            self.assertEqual(new_vim, self.vm_session.vim)


class TestConnectionHandler(base.TestCase):

    def setUp(self):
        super(TestConnectionHandler, self).setUp()
        self.host_ip = "192.168.1.3"
        self.host_username = "user"
        self.host_password = "password"
        self.api_retry_count = 2
        self.wsdl_url = "fake_url"

    def test_create_connection(self):
        vim_session.ConnectionHandler.set_vc_details(self.host_ip,
                                                     self.host_username,
                                                     self.host_password,
                                                     self.api_retry_count,
                                                     self.wsdl_url)
        vim_session.ConnectionHandler.create_session = False
        vm_session = vim_session.ConnectionHandler.create_connection()
        self.assertEqual(vim_session.ConnectionHandler.host_ip,
                         self.host_ip)
        self.assertEqual(vim_session.ConnectionHandler.host_username,
                         self.host_username)
        self.assertEqual(vim_session.ConnectionHandler.host_password,
                         self.host_password)
        self.assertEqual(vim_session.ConnectionHandler.api_retry_count, 2)
        self.assertEqual(vim_session.ConnectionHandler.wsdl_url,
                         self.wsdl_url)
        self.assertTrue(vm_session)

    def test_connection_handler_stop(self):
        vim_session.ConnectionHandler.set_vc_details(self.host_ip,
                                                     self.host_username,
                                                     self.host_password,
                                                     self.api_retry_count,
                                                     self.wsdl_url)
        vim_session.ConnectionHandler.create_session = False
        vim_session.ConnectionHandler.create_connection()
        with mock.patch.object(api.VMwareAPISession,
                               "logout") as log_ob:
            vim_session.ConnectionHandler.stop()
            self.assertTrue(log_ob.called)
            self.assertTrue(vim_session.ConnectionHandler.stopped)

    def test_connection_handler_start(self):
        vim_session.ConnectionHandler.start()
        self.assertFalse(vim_session.ConnectionHandler.stopped)

    def test_get_connection(self):
        vim_session.ConnectionHandler.set_vc_details(self.host_ip,
                                                     self.host_username,
                                                     self.host_password,
                                                     self.api_retry_count,
                                                     self.wsdl_url)
        vim_session.ConnectionHandler.create_session = False
        vm_session = vim_session.ConnectionHandler.create_connection()
        new_session = vim_session.ConnectionHandler.get_connection()
        self.assertEqual(vm_session, new_session)



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log
from oslo_utils import importutils

from neutron.agent import securitygroups_rpc as sg_rpc
import time

LOG = log.getLogger(__name__)


class OVSVAppSecurityGroupAgent(sg_rpc.SecurityGroupAgentRpc):
    """OVSvApp derived class from OVSSecurityGroupAgent

    This class is to override the default behavior of some methods.
    """
    def __init__(self, context, plugin_rpc, defer_apply):
        self.context = context
        self.plugin_rpc = plugin_rpc
        self.init_firewall(defer_apply)
        LOG.info(_("OVSVAppSecurityGroupAgent initialized."))

    def init_firewall(self, defer_refresh_firewall=False):
        firewall_driver = cfg.CONF.SECURITYGROUP.ovsvapp_firewall_driver
        LOG.debug("Init firewall settings (driver=%s).", firewall_driver)
        if not firewall_driver:
            firewall_driver = 'neutron.agent.firewall.NoopFirewallDriver'
        self.firewall = importutils.import_object(firewall_driver)
        # The following flag will be set to true if port filter must not be
        # applied as soon as a rule or membership notification is received.
        self.defer_refresh_firewall = defer_refresh_firewall
        # Stores devices for which firewall should be refreshed when
        # deferred refresh is enabled.
        self.devices_to_refilter = set()
        # Flag raised when a global refresh is needed.
        self.global_refresh_firewall = False
        self._use_enhanced_rpc = None

    def add_devices_to_filter(self, devices):
        if not devices:
            return
        self.firewall.add_ports_to_filter(devices)

    def ovsvapp_sg_update(self, port_with_rules):
        for port_id in port_with_rules:
            if port_id in self.firewall.ports:
                self.firewall.prepare_port_filter(port_with_rules[port_id])

    def remove_device_filters(self, device_id):
        if not device_id:
            return
        LOG.info(_("Remove device filters for %r."), device_id)
        self.firewall.clean_port_filters([device_id], True)

    def prepare_firewall(self, device_ids):
        """Puts in new rules for input port_ids.

        This routine puts in new rules for the
        input ports shippped as device_ids.

        :param device_ids: set of port_ids for which firewall rules
        need to be created.
        """
        LOG.info(_("Prepare firewall rules %s."), len(device_ids))
        dev_list = list(device_ids)
        if len(dev_list) > 10:
            sublists = [dev_list[x:x + 10] for x in xrange(0, len(dev_list),
                                                           10)]
        else:
            sublists = [dev_list]
        for dev_ids in sublists:
            devices = self.plugin_rpc.security_group_rules_for_devices(
                self.context, dev_ids)
            for device in devices.values():
                if device['id'] in dev_ids:
                    self.firewall.prepare_port_filter(device)

    def refresh_firewall(self, device_ids=None):
        """Removes all rules for input port_ids and puts in new ruels for them.

        This routine erases all rules and puts in new rules for the
        input ports shippped as device_ids.

        :param device_ids: set of port_ids for which firewall rules
        need to be refreshed.
        """
        LOG.info(_("Refresh firewall rules."))
        if not device_ids:
            device_ids = self.firewall.ports.keys()
            if not device_ids:
                LOG.info(_("No ports here to refresh firewall."))
                return
        dev_list = list(device_ids)
        if len(dev_list) > 10:
            sublists = [dev_list[x:x + 10] for x in xrange(0, len(dev_list),
                                                           10)]
        else:
            sublists = [dev_list]

        for dev_ids in sublists:
            # Sleep is to prevent any device_create calls from getting starved.
            time.sleep(0)
            devices = self.plugin_rpc.security_group_rules_for_devices(
                self.context, dev_ids)
            for device in devices.values():
                if device['id'] in dev_ids:
                    self.firewall.update_port_filter(device)

    def refresh_port_filters(self, own_devices, other_devices):
        """Update port filters for devices.

        This routine refreshes firewall rules when devices have been
        updated, or when there are changes in security group membership
        or rules.

        :param own_devices: set containing identifiers for devices
        belonging to this ESX host.
        :param other_devices: set containing identifiers for
        devices belonging to other ESX hosts within the Cluster.
        """
        # These data structures are cleared here in order to avoid
        # losing updates occurring during firewall refresh.
        devices_to_refilter = self.devices_to_refilter
        global_refresh_firewall = self.global_refresh_firewall
        self.devices_to_refilter = set()
        self.global_refresh_firewall = False
        LOG.info(_("Going to refresh for devices: %s."),
                 devices_to_refilter)
        if global_refresh_firewall:
            LOG.debug("Refreshing firewall for all filtered devices.")
            self.firewall.clean_port_filters(other_devices)
            self.refresh_firewall()
        else:
            own_devices = (own_devices & devices_to_refilter)
            other_devices = (other_devices & devices_to_refilter)
            self.firewall.clean_port_filters(other_devices)
            if own_devices:
                LOG.info(_("Refreshing firewall for %d devices."),
                         len(own_devices))
                self.refresh_firewall(own_devices)
            if other_devices:
                LOG.info(_("Refreshing firewall for %d devices."),
                         len(other_devices))
                self.prepare_firewall(other_devices)



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import contextlib

import fixtures
import mock

from networking_vsphere.common import constants
from networking_vsphere.common import error
from networking_vsphere.common import model
from networking_vsphere.drivers import vc_driver as vmware_driver
from networking_vsphere.tests import base
from networking_vsphere.tests.unit.utils import fake_vmware_api
from networking_vsphere.tests.unit.utils import stubs
from networking_vsphere.utils import cache
from networking_vsphere.utils import resource_util
from networking_vsphere.utils import vim_util

VcCache = cache.VCCache


def fake_is_valid_switch(obj, cluster_mor, switch):
    return fake_vmware_api._db_content["HostSystem"].values()


def fake_get_unused_portgroups(obj, switch):
    return []


def fake_delete_portgroup(obj, switch, pg):
    return


def fake_create_network(obj, network, virtual_switch):
    return


class TestVmwareDriver(base.TestCase):

    def setUp(self):
        super(TestVmwareDriver, self).setUp()
        self.cluster_dvs_mapping = {"ClusterComputeResource": "test_dvs"}
        self.fake_visdk = self.useFixture(stubs.FakeVmware())
        self.session = self.fake_visdk.session
        self.useFixture(stubs.CacheFixture())
        self.useFixture(fixtures.MonkeyPatch(
            'networking_vsphere.drivers.vc_driver.'
            'VCNetworkDriver.is_valid_switch', fake_is_valid_switch))
        self.useFixture(fixtures.MonkeyPatch(
            'networking_vsphere.drivers.vc_driver.'
            'VCNetworkDriver.get_unused_portgroups',
            fake_get_unused_portgroups))
        self.useFixture(fixtures.MonkeyPatch(
            'networking_vsphere.drivers.vc_driver.'
            'VCNetworkDriver.delete_portgroup', fake_delete_portgroup))
        self.useFixture(fixtures.MonkeyPatch(
            'networking_vsphere.drivers.vc_driver.'
            'VCNetworkDriver.create_network', fake_create_network))
        self.vc_driver = vmware_driver.VCNetworkDriver()
        self.vc_driver.state = constants.DRIVER_RUNNING
        self.vc_driver.add_cluster("ClusterComputeResource", "test_dvs")

    def test_stop(self):
        with mock.patch.object(vim_util, "cancel_wait_for_updates",
                               return_value=None):
            self.vc_driver.stop()
            self.assertEqual(self.vc_driver.state, constants.DRIVER_STOPPED)

    def test_add_cluster_none(self):
        old_mapping = cache.VCCache.get_cluster_switch_mapping()
        self.vc_driver.add_cluster("", "test_dvs")
        self.assertEqual(old_mapping,
                         cache.VCCache.get_cluster_switch_mapping(),
                         "Cluster mapping got changed even for invalid"
                         "mapping")

    def test_add_cluster_invalid_switch(self):
        with mock.patch.object(self.vc_driver, "is_valid_switch",
                               return_value=None):
            self.assertIn("ClusterComputeResource",
                          cache.VCCache.get_cluster_switch_mapping())
            self.vc_driver.add_cluster("ClusterComputeResource", "invalid_dvs")
            self.assertNotIn("ClusterComputeResource",
                             cache.VCCache.get_cluster_switch_mapping())

    def test_add_cluster_invalid_cluster_path(self):
        self.vc_driver.add_cluster("invalid_cluster", "test_dvs")
        self.assertNotIn("invalid_cluster",
                         cache.VCCache.get_cluster_switch_mapping())

    def test_add_cluster_updatevds(self):
        self.assertEqual(cache.VCCache.
                         get_switch_for_cluster_path(
                             "ClusterComputeResource"), "test_dvs")
        self.vc_driver.add_cluster("ClusterComputeResource", "new_dvs")
        self.assertEqual(cache.VCCache.
                         get_switch_for_cluster_path(
                             "ClusterComputeResource"), "new_dvs")

    def test_add_cluster_clusterchanged(self):
        self.vc_driver.state = constants.DRIVER_IDLE
        cluster_mor = resource_util.get_cluster_mor_by_path(
            self.session, "ClusterComputeResource")
        old_clu_id = cluster_mor.value
        object.__setattr__(cluster_mor, 'value', "new_value")
        with contextlib.nested(
            mock.patch.object(resource_util, "get_cluster_mor_by_path",
                              return_value=cluster_mor),
            mock.patch.object(self.vc_driver,
                              "_unregister_cluster_for_updates",
                              return_value=None)):
                self.assertEqual(cache.VCCache.get_switch_for_cluster_path(
                    "ClusterComputeResource"), "test_dvs")
                self.assertIn(old_clu_id,
                              cache.VCCache.cluster_id_to_path)
                self.vc_driver.add_cluster("ClusterComputeResource", "new_dvs")
                self.assertNotIn(old_clu_id,
                                 cache.VCCache.cluster_id_to_path)
                self.assertIn("new_value",
                              cache.VCCache.cluster_id_to_path)
                self.assertEqual(cache.VCCache.get_switch_for_cluster_path(
                    "ClusterComputeResource"), "new_dvs")
                self.assertEqual(self.vc_driver.state, constants.DRIVER_READY)

    def test_is_connected_none(self):
        self.vc_driver.session = None
        self.assertFalse(self.vc_driver.is_connected())

    def test_create_port(self):
        vm_id = fake_vmware_api.Constants.VM_UUID
        vlan = model.Vlan(vlanIds=["1001"])
        network_config = model.NetworkConfig(vlan)
        network = model.Network(
            name="net-1234", network_type=constants.NETWORK_VLAN,
            config=network_config)
        port = model.Port(name=None,
                          mac_address=None,
                          ipaddresses=None,
                          vm_id=vm_id)
        virtual_nic = model.VirtualNic(mac_address=None,
                                       port_uuid=None,
                                       vm_id=vm_id,
                                       vm_name=None,
                                       nic_type=None)
        with mock.patch.object(model, "VirtualSwitch") as vswitch:
            self.vc_driver.create_port(network, port, virtual_nic)
            self.assertTrue(vswitch.called)

    def test_create_port_exc(self):
        vm_id = fake_vmware_api.Constants.VM_UUID
        vlan = model.Vlan(vlanIds=["1001"])
        network_config = model.NetworkConfig(vlan)
        network = model.Network(
            name="net-1234", network_type=constants.NETWORK_VLAN,
            config=network_config)
        port = model.Port(name=None,
                          mac_address=None,
                          ipaddresses=None,
                          vm_id=vm_id)
        virtual_nic = model.VirtualNic(mac_address=None,
                                       port_uuid=None,
                                       vm_id=vm_id,
                                       vm_name=None, nic_type=None)
        with mock.patch.object(self.vc_driver, "is_valid_switch",
                               return_value=None):
            exc = self.assertRaises(error.VcenterConfigurationError,
                                    self.vc_driver.create_port,
                                    network, port, virtual_nic)
            self.assertIn("Invalid Switch", str(exc))

    def test_create_port_invalid_cluster(self):
        vm_id = fake_vmware_api.Constants.VM_UUID
        vlan = model.Vlan(vlanIds=["1001"])
        network_config = model.NetworkConfig(vlan)
        network = model.Network(
            name="net-1234", network_type=constants.NETWORK_VLAN,
            config=network_config)
        port = model.Port(name=None,
                          mac_address=None,
                          ipaddresses=None,
                          vm_id=vm_id)
        virtual_nic = model.VirtualNic(mac_address=None,
                                       port_uuid=None,
                                       vm_id=vm_id,
                                       vm_name=None,
                                       nic_type=None)
        cluster_mor = fake_vmware_api.DataObject()
        cluster_mor.value = "invalid_id"
        cache.VCCache.add_cluster_mor_for_vm(vm_id, cluster_mor)
        exc = self.assertRaises(error.VcenterConfigurationError,
                                self.vc_driver.create_port,
                                network, port, virtual_nic)
        self.assertIn("Cluster for VM %s could not be determined" %
                      vm_id, str(exc))

    def test_process_update_set_filterset_none(self):
        updateSet = fake_vmware_api.DataObject()
        updateSet.version = 1
        updateSet.filterSet = None
        events = self.vc_driver._process_update_set(updateSet)
        self.assertEqual(len(events), 0)

    def test_process_update_set_objectset_none(self):
        updateSet = fake_vmware_api.DataObject()
        updateSet.version = 1
        filterSet = []
        updateSet.filterSet = filterSet
        propFilterUpdate = fake_vmware_api.DataObject()
        filterSet.append(propFilterUpdate)
        propFilterUpdate.objectSet = None
        events = self.vc_driver._process_update_set(updateSet)
        self.assertEqual(len(events), 0)

    def test_process_update_set_invalid(self):
        updateSet = fake_vmware_api.DataObject()
        updateSet.version = 1
        filterSet = []
        updateSet.filterSet = filterSet
        propFilterUpdate = fake_vmware_api.DataObject()
        filterSet.append(propFilterUpdate)
        objectSet = []
        propFilterUpdate.objectSet = objectSet
        objectUpdate = fake_vmware_api.DataObject()
        objectUpdate.obj = (
            fake_vmware_api._db_content["ClusterComputeResource"].values()[0])
        objectUpdate.kind = "enter"
        changeSet = []
        objectUpdate.changeSet = changeSet
        for prop in objectUpdate.obj.propSet:
            changeSet.append(prop)
        objectSet.append(objectUpdate)
        events = self.vc_driver._process_update_set(updateSet)
        self.assertEqual(len(events), 0)

    def test_process_update_set_modify(self):
        updateSet = fake_vmware_api.DataObject()
        updateSet.version = 1
        filterSet = []
        updateSet.filterSet = filterSet
        propFilterUpdate = fake_vmware_api.DataObject()
        filterSet.append(propFilterUpdate)
        objectSet = []
        propFilterUpdate.objectSet = objectSet
        objectUpdate = fake_vmware_api.DataObject()
        objectUpdate.obj = (
            fake_vmware_api._db_content["VirtualMachine"].values()[0])
        objectUpdate.kind = "modify"
        changeSet = []
        objectUpdate.changeSet = changeSet
        for prop in objectUpdate.obj.propSet:
            if prop.name == "runtime.host":
                delattr(prop, "val")
            changeSet.append(prop)
        objectSet.append(objectUpdate)
        events = self.vc_driver._process_update_set(updateSet)
        self.assertEqual(len(events), 1)
        self.assertEqual(events[0].event_type, constants.VM_CREATED)

    def test_process_update_set_leave(self):
        updateSet = fake_vmware_api.DataObject()
        updateSet.version = 1
        filterSet = []
        updateSet.filterSet = filterSet
        propFilterUpdate = fake_vmware_api.DataObject()
        filterSet.append(propFilterUpdate)
        objectSet = []
        propFilterUpdate.objectSet = objectSet
        objectUpdate = fake_vmware_api.DataObject()
        objectUpdate.obj = (
            fake_vmware_api._db_content["VirtualMachine"].values()[0])
        objectUpdate.kind = "leave"
        changeSet = []
        objectUpdate.changeSet = changeSet
        objectSet.append(objectUpdate)
        events = self.vc_driver._process_update_set(updateSet)
        self.assertEqual(len(events), 0)

    def test_process_update_set_invalid_extraConfig(self):
        updateSet = fake_vmware_api.DataObject()
        updateSet.version = 1
        filterSet = []
        updateSet.filterSet = filterSet
        propFilterUpdate = fake_vmware_api.DataObject()
        filterSet.append(propFilterUpdate)
        objectSet = []
        propFilterUpdate.objectSet = objectSet
        objectUpdate = fake_vmware_api.DataObject()
        objectUpdate.obj = (
            fake_vmware_api._db_content["VirtualMachine"].values()[0])
        objectUpdate.kind = "modify"
        changeSet = []
        objectUpdate.changeSet = changeSet
        for prop in objectUpdate.obj.propSet:
            if prop.name == 'config.extraConfig["nvp.vm-uuid"]':
                delattr(prop, "val")
            changeSet.append(prop)
        objectSet.append(objectUpdate)
        events = self.vc_driver._process_update_set(updateSet)
        self.assertEqual(len(events), 0)

    def test_delete_stale_portgroups(self):
        with contextlib.nested(
            mock.patch.object(
                self.vc_driver,
                "get_unused_portgroups",
                return_value=[fake_vmware_api.Constants.PORTGROUP_NAME]),
            mock.patch.object(self.vc_driver, "delete_portgroup")
            ) as (unused_ob, delete_pg_ob):
            self.vc_driver.delete_stale_portgroups("test_dvs")
            self.assertTrue(unused_ob.called)
            self.assertTrue(delete_pg_ob.called)

    def test_post_delete_vm(self):
        uuid = fake_vmware_api.Constants.VM_UUID
        clus_mor = (
            fake_vmware_api._db_content["ClusterComputeResource"].values()[0])
        vm_mor = fake_vmware_api._db_content["VirtualMachine"].values()[0]
        VcCache.add_cluster_mor_for_vm(uuid, clus_mor)
        VcCache.add_vm_mor_for_uuid(uuid, vm_mor)
        vm_model = model.VirtualMachine(name=vm_mor.name,
                                        vnics=[],
                                        uuid=uuid,
                                        key=vm_mor.value)
        VcCache.add_vm_model_for_uuid(uuid, vm_model)
        self.assertIn(uuid, VcCache.vm_to_cluster)
        self.assertIn(uuid, VcCache.vm_uuid_to_mor)
        self.assertIn(vm_mor.value, VcCache.vm_moid_to_uuid)
        self.assertIn(uuid, VcCache.vm_uuid_to_model)
        self.vc_driver.post_delete_vm(vm_model)
        self.assertNotIn(uuid, VcCache.vm_to_cluster)
        self.assertNotIn(uuid, VcCache.vm_uuid_to_mor)
        self.assertNotIn(vm_mor.value, VcCache.vm_moid_to_uuid)
        self.assertNotIn(uuid, VcCache.vm_uuid_to_model)

    def test_remove_cluster(self):
        with mock.patch.object(self.vc_driver,
                               "_unregister_cluster_for_updates"
                               ) as unreg_ob:
            self.vc_driver.remove_cluster("ClusterComputeResource",
                                          "test_dvs")
            self.assertTrue(unreg_ob.called)

    def test_remove_cluster_invalid_cluster_path(self):
        with contextlib.nested(
            mock.patch.object(self.vc_driver,
                              "_unregister_cluster_for_updates"),
            mock.patch.object(self.vc_driver,
                              "_find_cluster_id_for_path",
                              return_value="1234")
            ) as (unreg_ob, find_ob):
            self.vc_driver.remove_cluster("invalid_cluster",
                                          "test_dvs")
            self.assertFalse(unreg_ob.called)
            self.assertFalse(find_ob.called)

    def test_remove_cluster_id_none(self):
        with contextlib.nested(
            mock.patch.object(self.vc_driver,
                              "_unregister_cluster_for_updates"),
            mock.patch.object(self.vc_driver,
                              "_find_cluster_id_for_path",
                              return_value=None)
            ) as (unreg_ob, find_ob):
            self.vc_driver.remove_cluster("ClusterComputeResource",
                                          "test_dvs")
            self.assertFalse(unreg_ob.called)
            self.assertTrue(find_ob.called)



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_vsphere.tests import base
from networking_vsphere.tests.unit.utils import fake_vmware_api
from networking_vsphere.tests.unit.utils import stubs
from networking_vsphere.utils import cache


class VCCacheTestCase(base.TestCase):

    def setUp(self):
        base.TestCase.setUp(self)
        self.useFixture(stubs.CacheFixture())

    def test_add_esx_hostname_for_vm(self):
        vm_uuid = "VM-1234-5678"
        hostname = "fake_esx_host"
        cache.VCCache.add_esx_hostname_for_vm(vm_uuid, hostname)
        self.assertEqual(hostname,
                         cache.VCCache.get_esx_hostname_for_vm(vm_uuid))

    def test_add_cluster_mor_for_vm(self):
        vm_uuid = "VM-1234-5678-%s"
        cluster_mor = "Cluster-1234-5678-%s"
        for i in range(0, 1000):
            cache.VCCache.add_cluster_mor_for_vm(vm_uuid % i, cluster_mor % i)
        cache.VCCache.add_cluster_mor_for_vm(vm_uuid % 1000,
                                             cluster_mor % 1000)
        self.assertEqual(1000, len(cache.VCCache.vm_to_cluster))
        self.assertEqual(cluster_mor % 1000,
                         cache.VCCache.get_cluster_mor_for_vm(vm_uuid % 1000))

    def test_add_path_for_cluster_id(self):
        cluster_id = "Cluster-1234-5678"
        cluster_path = "fake_path"
        cache.VCCache.add_path_for_cluster_id(cluster_id, cluster_path)
        self.assertEqual(cluster_path,
                         cache.VCCache.get_cluster_path_for_id(cluster_id))

    def test_get_cluster_path_for_id_notexisting(self):
        cluster_id = "Cluster-1234-5678"
        self.assertIsNone(cache.VCCache.get_cluster_path_for_id(cluster_id))

    def test_add_vm_mor_for_uuid(self):
        uuid = "VM-1234-5678"
        vm_mor = fake_vmware_api.DataObject()
        vm_mor.value = "vm-123"
        cache.VCCache.add_vm_mor_for_uuid(uuid, vm_mor)
        self.assertEqual(vm_mor, cache.VCCache.get_vm_mor_for_uuid(uuid))

    def test_add_vm_model_for_uuid(self):
        vm_uuid = "VM-1234-5678"
        vm = "VirtualMachine"
        cache.VCCache.add_vm_model_for_uuid(vm_uuid, vm)
        self.assertEqual(vm,
                         cache.VCCache.get_vm_model_for_uuid(vm_uuid))

    def test_add_switch_for_cluster_path(self):
        cluster_path = "fake_path"
        switch_name = "fake_switch"
        cache.VCCache.add_switch_for_cluster_path(cluster_path, switch_name)
        self.assertEqual(switch_name,
                         cache.VCCache.get_switch_for_cluster_path(
                             cluster_path))

    def test_remove_vm_for_uuid(self):
        uuid = "VM-1234-5678"
        vm_mor = fake_vmware_api.DataObject()
        vm_mor.value = "vm-123"
        cache.VCCache.add_vm_mor_for_uuid(uuid, vm_mor)
        self.assertEqual(cache.VCCache.get_vm_mor_for_uuid(uuid),
                         vm_mor)
        self.assertEqual(cache.VCCache.get_vmuuid_for_moid(vm_mor.value),
                         uuid)

        cache.VCCache.remove_vm_for_uuid(uuid)
        self.assertIsNone(cache.VCCache.get_vm_mor_for_uuid(uuid))
        self.assertIsNone(cache.VCCache.get_vmuuid_for_moid(vm_mor.value))

    def test_remove_cluster_path(self):
        cluster_path = "fake_path"
        switch_name = "fake_switch"
        cache.VCCache.add_switch_for_cluster_path(cluster_path, switch_name)

        cache.VCCache.remove_cluster_path(cluster_path)
        self.assertIsNone(cache.VCCache.get_switch_for_cluster_path(
                          cluster_path))

    def test_remove_cluster_id(self):
        cluster_id = "Cluster-1234-5678"
        cluster_path = "fake_path"
        cache.VCCache.add_path_for_cluster_id(cluster_id, cluster_path)

        cache.VCCache.remove_cluster_id(cluster_id)
        self.assertIsNone(cache.VCCache.get_cluster_path_for_id(cluster_id))

    def test_get_cluster_id_for_path(self):
        cluster_id = "Cluster-1234-5678"
        cluster_path = "fake_path"
        cache.VCCache.add_path_for_cluster_id(cluster_id, cluster_path)

        self.assertEqual(cluster_id,
                         cache.VCCache.get_cluster_id_for_path(cluster_path))

    def test_get_cluster_id_for_path_notexisting(self):
        cluster_path = "fake_path"
        self.assertIsNone(cache.VCCache.get_cluster_id_for_path(cluster_path))

    def test_get_cluster_switch_mapping(self):
        cluster_path = "fake_path"
        switch_name = "fake_switch"
        cache.VCCache.add_switch_for_cluster_path(cluster_path, switch_name)
        # TODO(romilg): Revisit to use assertEqual here.
        self.assertIsNotNone(cache.VCCache.get_cluster_switch_mapping)


# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import re
import time

from oslo_log import log

from networking_vsphere.common import constants
from networking_vsphere.common import error
from networking_vsphere.common import model
from networking_vsphere.common import utils
from networking_vsphere.drivers import driver
from networking_vsphere.utils import cache
from networking_vsphere.utils import common_util
from networking_vsphere.utils import error_util
from networking_vsphere.utils import network_util
from networking_vsphere.utils import resource_util
from networking_vsphere.utils import vim_session
from networking_vsphere.utils import vim_util

LOG = log.getLogger(__name__)


class VCNetworkDriver(driver.NetworkDriver):

    def __init__(self):
        driver.NetworkDriver.__init__(self)
        self.state = constants.DRIVER_IDLE
        self.clusters_by_id = {}
        self.cluster_id_to_filter = {}
        cache.VCCache.reset()
        self.session = vim_session.ConnectionHandler.get_connection()

    def get_unused_portgroups(self, switch):
        raise NotImplementedError()

    def delete_portgroup(self, switch, pg):
        raise NotImplementedError()

    def is_valid_switch(self, cluster_mor, switch):
        """Validate if the switch is valid for this cluster.

           Makes sure that the switch is present on all the hosts of the
           cluster and returns list of hosts present on the cluster,
           if switch is valid else returns None.
        """
        raise NotImplementedError()

    def delete_stale_portgroups(self, switch):
        LOG.info(_("Deleting unused portgroups on %s."), switch)
        port_group_names = self.get_unused_portgroups(switch)
        uuid_regex = ("[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}"
                      "-[0-9a-f]{4}-[0-9a-f]{12}")
        for port_group in port_group_names:
            if re.match(uuid_regex, port_group, re.IGNORECASE):
                self.delete_portgroup(switch, port_group)

    def validate_cluster_switch_mapping(self, cluster_path, switch):
        """Validate the cluster_switch_mapping."""
        if not cluster_path or not switch:
            return False, None
        cluster_mor = resource_util.get_cluster_mor_by_path(self.session,
                                                            cluster_path)
        if not cluster_mor:
            LOG.error(_("Invalid cluster: %s."), cluster_path)
            return False, None
        else:
            if not self.is_valid_switch(cluster_mor, switch):
                LOG.error(_("Invalid Switch: %(switch)s for cluster: "
                            "%(path)s."),
                          {'switch': switch, 'path': cluster_path})
                return False, None
            else:
                LOG.info(_("Cluster: %(path)s and switch: %(sw)s are "
                           "validated."),
                         {'path': cluster_path, 'sw': switch})
                return True, cluster_mor

    def _find_cluster_id_for_path(self, path):
        return cache.VCCache.get_cluster_id_for_path(path)

    def _unregister_cluster_for_updates(self, cluster_mor):
        property_filter_obj = self.cluster_id_to_filter[cluster_mor.value]
        if property_filter_obj:
            self.session._call_method(self.session._get_vim(),
                                      "DestroyPropertyFilter",
                                      property_filter_obj)

    def _register_cluster_for_updates(self, cluster_mor):
        vm_properties = ['name',
                         'config.extraConfig["nvp.vm-uuid"]',
                         'runtime.host',
                         'config.hardware.device']
        propertyDict = {"VirtualMachine": vm_properties}
        property_filter_spec = self.session._call_method(
            vim_util,
            "get_property_filter_specs",
            propertyDict,
            [cluster_mor])
        property_filter_obj = self.session._call_method(vim_util,
                                                        "create_filter",
                                                        property_filter_spec)
        return property_filter_obj

    def _find_cluster_switch_for_vm(self, device_id):
        cluster_mor = resource_util.get_cluster_mor_for_vm(self.session,
                                                           device_id)
        cluster_path = None
        if cluster_mor:
            cluster_id = cluster_mor.value
            cluster_path = cache.VCCache.get_cluster_path_for_id(cluster_id)
        if not cluster_path:
            raise error.VcenterConfigurationError(_("Cluster for VM %s could "
                                                  "not be determined")
                                                  % device_id)
        switch_name = cache.VCCache.get_switch_for_cluster_path(cluster_path)
        return cluster_mor, cluster_path, switch_name

    def add_cluster(self, cluster_path, switch_name):
        LOG.info(_("Adding cluster_switch_mapping %(path)s:%(switch)s."),
                 {'path': cluster_path, 'switch': switch_name})
        if (cluster_path in cache.VCCache.get_cluster_switch_mapping() and
            cache.VCCache.get_switch_for_cluster_path(
                cluster_path) == switch_name):
            LOG.info(_("cluster_switch_mapping %(cp)s:%(sw)s already "
                       "present."),
                     {'cp': cluster_path, 'sw': switch_name})
            return
        valid, cluster_mor = self.validate_cluster_switch_mapping(cluster_path,
                                                                  switch_name)
        if not valid:
            if cluster_path in cache.VCCache.get_cluster_switch_mapping():
                LOG.info(_("Removing invalid cluster_switch_mapping: "
                         "%(cp)s:%(sw)s."),
                         {'cp': cluster_path, 'sw': switch_name})
                self.remove_cluster(cluster_path, switch_name)
            return
        if cluster_path in cache.VCCache.get_cluster_switch_mapping():
            cluster_id = self._find_cluster_id_for_path(cluster_path)
            if cluster_id == cluster_mor.value:
                LOG.info(_("Updating switch name for cluster: "
                         "%(cp)s to %(sw)s."),
                         {'cp': cluster_path, 'sw': switch_name})
                cache.VCCache.add_switch_for_cluster_path(cluster_path,
                                                          switch_name)
                self.delete_stale_portgroups(switch_name)
                return
            else:
                # Now this path points to a different cluster.
                LOG.info(_("Removing cluster %(cid)s as now path %(cp)s "
                           "points to a different cluster."),
                         {'cid': cluster_id, 'cp': cluster_path})
                old_clu_mor = self.clusters_by_id[cluster_id]
                if cluster_id in self.cluster_id_to_filter:
                    self._unregister_cluster_for_updates(old_clu_mor)
                    del self.cluster_id_to_filter[cluster_id]
                cache.VCCache.remove_cluster_id(cluster_id)
                if cluster_id in self.clusters_by_id:
                    del self.clusters_by_id[cluster_id]
        LOG.info(_("Registering cluster for mapping %(cp)s:%(sw)s."),
                 {'cp': cluster_path, 'sw': switch_name})
        property_filter_obj = self._register_cluster_for_updates(cluster_mor)
        # Cache the cluster.
        cache.VCCache.add_path_for_cluster_id(cluster_mor.value, cluster_path)
        self.clusters_by_id[cluster_mor.value] = cluster_mor
        self.cluster_id_to_filter[cluster_mor.value] = property_filter_obj
        cache.VCCache.add_switch_for_cluster_path(cluster_path,
                                                  switch_name)
        self.delete_stale_portgroups(switch_name)
        if self.state != constants.DRIVER_RUNNING and self.is_connected():
            self.state = constants.DRIVER_READY

    def remove_cluster(self, cluster_path, switch_name):
        mapping = "%s:%s" % (cluster_path, switch_name)
        LOG.info(_("Removing cluster_switch_mapping: %s."), mapping)
        if cluster_path not in cache.VCCache.get_cluster_switch_mapping():
            LOG.info(_("cluster_switch_mapping %s not present."), mapping)
            return
        cluster_id = self._find_cluster_id_for_path(cluster_path)
        if not cluster_id:
            LOG.info(_("Cluster for cluster_switch_mapping %s "
                       "not present in cache."), mapping)
        else:
            LOG.info(_("Unregistering cluster for mapping: %s."), mapping)
            cluster_mor = self.clusters_by_id[cluster_id]
            if cluster_id in self.cluster_id_to_filter:
                self._unregister_cluster_for_updates(cluster_mor)
                del self.cluster_id_to_filter[cluster_id]
            cache.VCCache.remove_cluster_id(cluster_id)
        if cluster_id in self.clusters_by_id:
            del self.clusters_by_id[cluster_id]
        cache.VCCache.remove_cluster_path(cluster_path)
        if not cache.VCCache.get_cluster_switch_mapping():
            self.state = constants.DRIVER_IDLE

    def is_connected(self):
        if self.session:
            return True
        else:
            return False

    def stop(self):
        self.pause()
        self.session = None
        self.state = constants.DRIVER_STOPPED

    @utils.require_state(state=[constants.DRIVER_RUNNING], excp=False)
    def pause(self):
        self.session._call_method(vim_util, "cancel_wait_for_updates")
        self.state = constants.DRIVER_READY

    @utils.require_state(state=[constants.DRIVER_READY], excp=False)
    def monitor_events(self):
        try:
            LOG.info(_("Starting monitoring for vCenter updates"))
            version = ""
            self.state = constants.DRIVER_RUNNING
            while self.state in (constants.DRIVER_RUNNING):
                try:
                    LOG.debug("Waiting for vCenter updates...")
                    try:
                        updateSet = self.session._call_method(
                            vim_util,
                            "wait_for_updates_ex",
                            version)
                        if self.state != constants.DRIVER_RUNNING:
                            break
                    except error_util.SocketTimeoutException:
                        # Ignore timeout.
                        continue
                    if updateSet:
                        version = updateSet.version
                        events = self._process_update_set(updateSet)
                        LOG.debug("Sending events : %s.", events)
                        self.dispatch_events(events)
                except error_util.VimFaultException as e:
                    excp = e.exception_obj
                    # InvalidCollectorVersionFault happens
                    # on session re-connect.
                    # Re-initialize WaitForUpdatesEx.
                    if hasattr(excp.fault.detail,
                               "InvalidCollectorVersionFault"):
                        LOG.debug("InvalidCollectorVersionFault - "
                                  "Re-initializing vCenter updates "
                                  "monitoring.")
                        version = ""
                        for cluster_mor in self.clusters_by_id.values():
                            pfo = self._register_cluster_for_updates(
                                cluster_mor)
                            clu_id = cluster_mor.value
                            self.cluster_id_to_filter[clu_id] = pfo
                        continue
                except Exception:
                    LOG.exception(_("Exception while processing update set."))
                time.sleep(0)
            LOG.info(_("Stopped monitoring for vCenter updates."))
        except Exception:
            LOG.exception(_("Monitoring for vCenter updates failed."))

    def _process_update_set(self, updateSet):
        """Processes the updateSet and returns VM events."""

        events = []
        host_name = None
        clus_name = None
        clus_id = None
        LOG.debug("Processing UpdateSet version: %s.", updateSet.version)
        filterSet = updateSet.filterSet
        if not filterSet:
            return events
        for propFilterUpdate in filterSet:
            objectSet = propFilterUpdate.objectSet
            if not objectSet:
                continue
            for objectUpdate in objectSet:
                try:
                    obj_mor = objectUpdate.obj
                    if obj_mor._type != "VirtualMachine":
                        continue
                    if objectUpdate.kind == "enter":
                        event_type = constants.VM_CREATED
                    elif objectUpdate.kind == "modify":
                        event_type = constants.VM_UPDATED
                    elif objectUpdate.kind == "leave":
                        event_type = constants.VM_DELETED
                    else:
                        continue
                    host_changed = False
                    vm_uuid = None
                    changes = common_util.convert_objectupdate_to_dict(
                        objectUpdate)
                    if changes.get('config.extraConfig["nvp.vm-uuid"]'):
                        vm_uuid = changes.get('config.extraConfig'
                                              '["nvp.vm-uuid"]').value
                        event_type = constants.VM_CREATED
                        cache.VCCache.add_vm_mor_for_uuid(vm_uuid, obj_mor)
                    else:
                        vm_uuid = cache.VCCache.get_vmuuid_for_moid(
                            obj_mor.value)
                    if vm_uuid:
                        old_vm = cache.VCCache.get_vm_model_for_uuid(vm_uuid)
                        LOG.debug("Old VM: %s.", old_vm)
                        LOG.debug("cache.VCCache.vm_uuid_to_model: %s.",
                                  cache.VCCache.vm_uuid_to_model)
                        new_vm = None
                        if old_vm:
                            if event_type == constants.VM_CREATED:
                                # Our cache has information about VM. But event
                                # received is VM_CREATED. This means it is
                                # session restart case. So we should not add
                                # this to new event.
                                LOG.debug("Session restart event for VM %s",
                                          vm_uuid)
                                continue
                            new_vm = copy.deepcopy(old_vm)
                        else:
                            new_vm = model.VirtualMachine(name=None,
                                                          vnics=[],
                                                          uuid=None,
                                                          key=None)
                            LOG.debug("VM not found in cache. New created: "
                                      " %s.", new_vm)
                        new_vm.uuid = vm_uuid
                        new_vm.key = obj_mor.value
                        if changes.get('name'):
                            new_vm.name = changes.get('name')
                        if changes.get('runtime.host'):
                            # Host got changed / New VM.
                            clus_mor = self.session._call_method(
                                vim_util,
                                "get_dynamic_property",
                                changes.get('runtime.host'),
                                "HostSystem", "parent")
                            # Cache the VM and Cluster.
                            cache.VCCache.add_cluster_mor_for_vm(vm_uuid,
                                                                 clus_mor)
                        if event_type != constants.VM_DELETED:
                            extraconfigs = (
                                resource_util.get_extraconfigs_for_vm(
                                    self.session, obj_mor))
                            if changes.get('config.hardware.device'):
                                devices = changes.get('config.hardware.device')
                                nicdvs = network_util.get_vnics_from_devices(
                                    devices)
                                i = 0
                                vnics = []
                                for nicdev in nicdvs:
                                    macadd = nicdev.macAddress
                                    portid = extraconfigs.get("nvp.iface-id.%d"
                                                              % i)
                                    vnic = model.VirtualNic(
                                        mac_address=macadd,
                                        port_uuid=portid,
                                        vm_id=vm_uuid,
                                        vm_name=new_vm.name,
                                        nic_type=None,
                                        key=None)
                                    vnics.append(vnic)
                                    i += 1
                                new_vm.vnics = vnics
                            host_mor = resource_util.get_host_mor_for_vm(
                                self.session, vm_uuid)
                            clus_mor = resource_util.get_cluster_mor_for_vm(
                                self.session, vm_uuid)
                            host_name = (
                                resource_util.get_hostname_for_host_mor(
                                    self.session, host_mor))
                            old_host_name = (
                                cache.VCCache.get_esx_hostname_for_vm(vm_uuid))
                            if old_host_name and old_host_name != host_name:
                                host_changed = True
                            cache.VCCache.add_esx_hostname_for_vm(vm_uuid,
                                                                  host_name)
                            clus_name = (
                                resource_util.get_clustername_for_cluster_mor(
                                    self.session, clus_mor))
                            clus_id = (
                                resource_util.get_clusterid_for_cluster_mor(
                                    self.session, clus_mor))
                        elif event_type == constants.VM_DELETED:
                            host_name = cache.VCCache.get_esx_hostname_for_vm(
                                vm_uuid)
                        event = model.Event(event_type, new_vm, None,
                                            host_name, clus_name, clus_id,
                                            host_changed)
                        events.append(event)
                        cache.VCCache.add_vm_model_for_uuid(vm_uuid, new_vm)
                        LOG.debug("Added vm to cache: %s.", new_vm)
                        LOG.debug("cache.VCCache.vm_uuid_to_model: %s.",
                                  cache.VCCache.vm_uuid_to_model)
                    else:
                        LOG.debug("Ignoring update for VM: %s.",
                                  changes.get('name'))
                except Exception:
                    LOG.exception(_("Exception while processing update set "
                                    "for event %(event)s for vm %(vm)s."),
                                  {'event': event_type, 'vm': vm_uuid})
        LOG.debug("Finished processing UpdateSet version: %s.",
                  updateSet.version)
        return events

    @utils.require_state(state=[constants.DRIVER_READY,
                         constants.DRIVER_RUNNING])
    def create_port(self, network, port, virtual_nic):
        device_id = port.vm_id
        cluster_mor, cluster_path, switch = self._find_cluster_switch_for_vm(
            device_id)
        host_mors = self.is_valid_switch(cluster_mor, switch)
        if not host_mors:
            LOG.error(_("Invalid Switch: %(sw)s for cluster: %(cp)s."),
                      {'sw': switch, 'cp': cluster_path})
            raise error.VcenterConfigurationError("Invalid Switch: %s for "
                                                  "cluster: %s." %
                                                  (switch, cluster_path))
        hosts = []
        for host_mor in host_mors:
            hosts.append(model.Host(key=host_mor.value))
        vswitch = model.VirtualSwitch(switch, hosts=hosts)
        self.create_network(network, vswitch)

    @utils.require_state(state=[constants.DRIVER_READY,
                         constants.DRIVER_RUNNING])
    def post_delete_vm(self, vm):
        cache.VCCache.remove_vm_for_uuid(vm.uuid)



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import eventlet

from networking_vsphere.common import error


class NetworkDriver(object):

    '''Base Class for defining interface for all L2 network drivers.'''

    def __init__(self):
        # Reference to NetworkDriverCallback implementation.
        self.callback_impl = None

    def set_callback(self, callback_impl):
        '''Sets the implementation of NetworkDriverCallback.'''
        if not isinstance(callback_impl, NetworkDriverCallback):
            raise error.OVSvAppNeutronAgentError(
                "Invalid NetworkDriverCallback")
        self.callback_impl = callback_impl

    def monitor_events(self):
        '''Common model API - monitor for events.'''
        raise NotImplementedError()

    def pause(self):
        '''Driver will stop processing and go to waiting.'''
        pass

    def stop(self):
        """To be called when the process is shutting down.

        Implements any cleanups that are required.
        """
        pass

    def is_connected(self):
        """Represents the state of the driver.

        Whether its connected to hypervisor or not.
        """
        raise NotImplementedError()

    def create_network(self, network, virtual_switch):
        """Creates l2 network on the compute node.

        :param network: Type model.Network
        :param virtual_switch: Type model.VirtualSwitch
        """
        raise NotImplementedError()

    def delete_network(self, network, virtual_switch=None):
        """Deletes l2 network on the compute node.

        :param network: Type model.Network
        :param virtual_switch: Type model.VirtualSwitch
        """
        raise NotImplementedError()

    def update_port(self, network=None, port=None, virtual_nic=None):
        """Update the Port status UP/DOWN.

        :param network: Type model.Network
        :param port: Type model.Port
        :param virtual_nic: Type model.VirtualNic
        """
        raise NotImplementedError()

    def prepare_port_group(self, network, port, virtual_nic):
        """Prepares portgroup creation on DVS with specified configuration.

        Calls create_network if network does not exist.
        :param network: Type model.Network
        :param port: Type model.Port
        :param virtual_nic: Type model.VirtualNic
        """
        raise NotImplementedError()

    def update_port_group(self, network, port, virtual_nic):
        """Updates the portgroup on DVS.

        :param network: Type model.Network
        :param port: Type model.Port
        :param virtual_nic: Type model.VirtualNic
        """
        raise NotImplementedError()

    def get_vlanid_for_port_group(self, dvs_name, pg_name):
        '''Obtain VLAN id associated with a DVS portgroup.'''
        raise NotImplementedError()

    def get_vm_ref_by_uuid(self, vm_uuid):
        '''Obtain vm reference from uuid.'''
        raise NotImplementedError()

    def wait_for_portgroup(self, vm_ref, pg_name):
        '''Wait on a portgroup on a dvswitch for a vm.'''
        raise NotImplementedError()

    def process_delete_vm(self, vm):
        '''Post process for a VM_DELETED task.'''
        raise NotImplementedError()

    def dispatch_events(self, events):
        '''Dispatch events to the callback on different green threads.'''
        for event in events:
            eventlet.spawn_n(self.callback_impl.process_event, event)


class NetworkDriverCallback(object):

    '''Base class defining callback interface.'''

    def process_event(self, event):
        """Called when an event is detected on the hypervisor.

        :param event: Type model.Event
        """
        raise NotImplementedError()



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures

from networking_vsphere.tests.unit.utils import fake_vmware_api
from networking_vsphere.utils import cache
from networking_vsphere.utils import error_util
from networking_vsphere.utils import vim_session
from networking_vsphere.utils import vim_util


def fake_get_vim_object(arg):
    """Stubs out the VMwareAPISession's get_vim_object method."""
    return fake_vmware_api.FakeVim()


@classmethod
def fake_create_connection(cls):
    cls.session = FakeVMwareSession(cls.host_ip,
                                    cls.host_username,
                                    cls.host_password,
                                    cls.api_retry_count,
                                    cls.wsdl_url)
    return cls.session


class FakeVMwareSession(object):

    def __init__(self, host_ip, host_username, host_password,
                 api_retry_count, wsdl_url, scheme="https", https_port=443,
                 ca_cert=None):
        self._host_ip = host_ip
        self._host_username = host_username
        self._host_password = host_password
        self._https_port = https_port
        self.api_retry_count = api_retry_count
        self.wsdl_url = wsdl_url
        self._scheme = scheme
        self._session_id = None
        self.vim = fake_get_vim_object("fake_module")
        session = self.vim.Login(self.vim.get_service_content().sessionManager,
                                 userName=self._host_username,
                                 password=self._host_password)
        self._session_id = session.key

    def _get_vim(self):
        return self.vim

    def _call_method(self, module, method, *args, **kwargs):
        temp_module = module
        for method_elem in method.split("."):
            temp_module = getattr(temp_module, method_elem)
            return temp_module(self.vim, *args, **kwargs)

    def wait_for_task(self, task_ref):
        task_info = self._call_method(vim_util, "get_dynamic_property",
                                      task_ref, "Task", "info")
        if task_info.state == "error":
            raise error_util.RunTimeError("Incorrect Parameter")
        return


class FakeVmware(fixtures.Fixture):

    def __init__(self):
        self.session = None

    def setUp(self):
        super(FakeVmware, self).setUp()
        fake_vmware_api.reset()
        self.useFixture(fixtures.MonkeyPatch(
            'networking_vsphere.utils.vim_session.'
            'ConnectionHandler.create_connection', fake_create_connection))
        self.vcenter_ip = "192.168.1.3"
        self.vcenter_username = "user"
        self.vcenter_password = "password"
        self.api_retry_count = 2
        self.wsdl_loc = "https://192.168.1.3/sdk/fake.wsdl"
        vim_session.ConnectionHandler.set_vc_details(self.vcenter_ip,
                                                     self.vcenter_username,
                                                     self.vcenter_password,
                                                     self.api_retry_count,
                                                     self.wsdl_loc)
        self.session = vim_session.ConnectionHandler.create_connection()
        self.addCleanup(fake_vmware_api.cleanup)


class CacheFixture(fixtures.Fixture):

    def setUp(self):
        fixtures.Fixture.setUp(self)
        cache.VCCache.reset()
        self.addCleanup(cache.VCCache.reset)


# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import mock
from oslo_config import cfg

import contextlib
from neutron.agent.common import config
from neutron.agent.common import ovs_lib
from neutron.common import constants

from networking_vsphere.drivers import ovs_firewall as ovs_fw
from networking_vsphere.tests import base

fake_port = {'security_group_source_groups': 'abc',
             'mac_address': '00:11:22:33:44:55',
             'network_id': "netid",
             'id': "123",
             'security_groups': "abc",
             'lvid': "100",
             "security_group_rules": [
                 {"direction": "ingress",
                  "protocol": "tcp",
                  "port_range_min": 2001,
                  "port_range_max": 2009,
                  "source_port_range_min": 67,
                  "source_port_range_max": 68,
                  "ethertype": "IPv4",
                  "source_ip_prefix": "150.1.1.0/22",
                  "dest_ip_prefix": "170.1.1.0/22"}]}


class TestOVSFirewallDriver(base.TestCase):
    def setUp(self):
        super(TestOVSFirewallDriver, self).setUp()
        config.register_root_helper(cfg.CONF)
        cfg.CONF.set_override('security_bridge_mapping',
                              "br-fake:fake_if", 'SECURITYGROUP')
        with contextlib.nested(
                mock.patch('networking_vsphere.drivers.'
                           'ovs_firewall.OVSFirewallDriver.'
                           'setup_base_flows'),
                mock.patch('neutron.agent.common.ovs_lib.OVSBridge.'
                           'create'),
                mock.patch('neutron.agent.common.ovs_lib.OVSBridge.'
                           'set_secure_mode'),
                mock.patch('neutron.agent.common.ovs_lib.OVSBridge.'
                           'get_port_ofport',
                           return_value=5)):
            self.ovs_firewall = ovs_fw.OVSFirewallDriver()
            self.ovs_firewall.sg_br = mock.Mock()
            self.mock_br = ovs_lib.DeferredOVSBridge(self.ovs_firewall.sg_br)
            self.LOG = ovs_fw.LOG

    def test_get_compact_port(self):
        compact_port = {'security_group_source_groups': 'abc',
                        'mac_address': '00:11:22:33:44:55',
                        'network_id': "netid",
                        'id': "123",
                        'device': "123",
                        'security_groups': "abc",
                        'lvid': "100"}
        res = self.ovs_firewall._get_compact_port(fake_port)
        self.assertEqual(compact_port, res)

    def test_add_ovs_flow(self):
        with contextlib.nested(
            mock.patch.object(self.ovs_firewall.sg_br, 'deferred',
                              return_value=self.mock_br),
            mock.patch.object(self.mock_br, 'add_flow')
        ) as (deferred_fn, mock_add_flow):
            self.ovs_firewall._add_ovs_flow(self.mock_br, 0, 1, "normal")
            mock_add_flow.assert_called_with(priority=0, actions='normal',
                                             table=1)

    def test_add_ovs_flow_with_protocol(self):
        with contextlib.nested(
            mock.patch.object(self.ovs_firewall.sg_br, 'deferred',
                              return_value=self.mock_br),
            mock.patch.object(self.mock_br, 'add_flow')
        ) as (deferred_fn, mock_add_flow):
            # rule with protocol
            self.ovs_firewall._add_ovs_flow(self.mock_br, 0, 1, "normal",
                                            protocol="arp")
            mock_add_flow.assert_called_with(table=1, priority=0,
                                             proto="arp", actions="normal")

    def test_add_ovs_flow_with_dest_mac(self):
        with contextlib.nested(
            mock.patch.object(self.ovs_firewall.sg_br, 'deferred',
                              return_value=self.mock_br),
            mock.patch.object(self.mock_br, 'add_flow')
        ) as (deferred_fn, mock_add_flow):
            # rule with dl_dest
            dest_mac = "01:00:00:00:00:00"
            self.ovs_firewall._add_ovs_flow(self.mock_br, 0, 1, "normal",
                                            dl_dest=dest_mac)
            mock_add_flow.assert_called_with(table=1, priority=0,
                                             dl_dst=dest_mac,
                                             actions="normal")

    def test_add_ovs_flow_with_tcpflag(self):
        with contextlib.nested(
            mock.patch.object(self.ovs_firewall.sg_br, 'deferred',
                              return_value=self.mock_br),
            mock.patch.object(self.mock_br, 'add_flow')
        ) as (deferred_fn, mock_add_flow):
            # rule with tcp_flags
            t_flag = "+rst"
            self.ovs_firewall._add_ovs_flow(self.mock_br, 0, 1, "normal",
                                            tcp_flag=t_flag)
            mock_add_flow.assert_called_with(table=1, priority=0,
                                             proto=constants.PROTO_NAME_TCP,
                                             tcp_flags=t_flag,
                                             actions="normal")

    def test_add_ovs_flow_with_icmptype(self):
        with contextlib.nested(
            mock.patch.object(self.ovs_firewall.sg_br, 'deferred',
                              return_value=self.mock_br),
            mock.patch.object(self.mock_br, 'add_flow')
        ) as (deferred_fn, mock_add_flow):
            # rule with icmp_req_type
            self.ovs_firewall._add_ovs_flow(self.mock_br, 0, 1, "normal",
                                            icmp_req_type=11)
            mock_add_flow.assert_called_with(table=1, priority=0,
                                             proto=constants.PROTO_NAME_ICMP,
                                             icmp_type=11, actions="normal")

    def test_add_ports_to_filter(self):
        self.ovs_firewall.filtered_ports = {}
        self.ovs_firewall.add_ports_to_filter([fake_port])
        res_port = {'security_group_source_groups': 'abc',
                    'mac_address': '00:11:22:33:44:55',
                    'network_id': "netid",
                    'id': "123",
                    'security_groups': "abc",
                    'lvid': "100",
                    'device': "123"}
        self.assertIsNotNone(self.ovs_firewall.filtered_ports)
        ret_port = self.ovs_firewall.filtered_ports["123"]
        self.assertEqual(res_port, ret_port)

    def test_setup_aap_flows(self):
        port_with_app = copy.deepcopy(fake_port)
        key = "allowed_address_pairs"
        port_with_app[key] = [{'ip_address': '10.0.0.2',
                               'mac_address': 'aa:bb:cc:dd:ee:aa'},
                              {'ip_address': '10.0.0.3',
                               'mac_address': 'aa:bb:cc:dd:ee:ab'}]
        with contextlib.nested(
            mock.patch.object(self.ovs_firewall, '_get_port_vlan',
                              return_value=100),
            mock.patch.object(self.ovs_firewall.sg_br, 'deferred',
                              return_value=self.mock_br),
            mock.patch.object(self.mock_br, 'add_flow')
        ) as (get_vlan, deferred_fn, mock_add_flow):
            self.ovs_firewall._setup_aap_flows(self.mock_br, port_with_app)
            self.assertEqual(2, mock_add_flow.call_count)

    def test_setup_aap_flows_invalid_call(self):
        port_with_app = copy.deepcopy(fake_port)
        with contextlib.nested(
            mock.patch.object(self.ovs_firewall, '_get_port_vlan',
                              return_value=100),
            mock.patch.object(self.ovs_firewall.sg_br, 'deferred',
                              return_value=self.mock_br),
            mock.patch.object(self.mock_br, 'add_flow')
        ) as (get_vlan, deferred_fn, mock_add_flow):
            self.ovs_firewall._setup_aap_flows(self.mock_br, port_with_app)
            self.assertFalse(mock_add_flow.called)

    def test_get_net_prefix_len(self):
        ip_addr = "150.1.1.0/22"
        length = self.ovs_firewall._get_net_prefix_len(ip_addr)
        self.assertNotEqual(0, length)

        ip_addr = None
        length = self.ovs_firewall._get_net_prefix_len(ip_addr)
        self.assertEqual(0, length)

    def test_get_protocol(self):
        proto = self.ovs_firewall._get_protocol("IPv4", None)
        self.assertEqual(['ip'], proto)

        proto = self.ovs_firewall._get_protocol("IPv6", None)
        self.assertEqual(['ipv6'], proto)

        proto = self.ovs_firewall._get_protocol("IPv6", 'icmp')
        self.assertEqual(['icmp6'], proto)

        proto = self.ovs_firewall._get_protocol("IPv4", 'icmp')
        self.assertEqual(['icmp'], proto)

        proto = self.ovs_firewall._get_protocol("IPv4", 'udp')
        self.assertEqual(['udp'], proto)

        proto = self.ovs_firewall._get_protocol("IPv6", 'tcp')
        self.assertEqual(['tcp'], proto)

        proto = self.ovs_firewall._get_protocol("IPv6", 'unknown')
        self.assertEqual(['ipv6', 'unknown'], proto)

    def test_add_flow_with_range(self):
        flow = {"priority": 1}
        res_flow = {"priority": 1,
                    "tp_dst": 1,
                    "tp_src": 1}
        with contextlib.nested(
            mock.patch.object(self.ovs_firewall.sg_br, 'deferred',
                              return_value=self.mock_br),
            mock.patch.object(self.mock_br, 'add_flow')
        ) as (deferred_fn, mock_add_flow):
            self.ovs_firewall._add_flow_with_range(self.mock_br, flow,
                                                   1, 2, 1, 2)
            mock_add_flow.called_with(res_flow)
            self.assertEqual(4, mock_add_flow.call_count)

    def test_add_flow_with_multiple_range(self):
        flow = {"priority": 1}
        with contextlib.nested(
            mock.patch.object(self.ovs_firewall.sg_br, 'deferred',
                              return_value=self.mock_br),
            mock.patch.object(self.mock_br, 'add_flow')
        ) as (deferred_fn, mock_add_flow):
            self.ovs_firewall._add_flow_with_range(self.mock_br, flow,
                                                   1, 3, 1, 2)
            self.assertEqual(6, mock_add_flow.call_count)

    def test_add_flows_call_no_vlan(self):
        port_with_app = copy.deepcopy(fake_port)
        with contextlib.nested(
            mock.patch.object(self.ovs_firewall, '_get_port_vlan',
                              return_value=None),
            mock.patch.object(self.ovs_firewall.sg_br, 'deferred',
                              return_value=self.mock_br),
            mock.patch.object(self.mock_br, 'add_flow'),
            mock.patch.object(self.LOG, 'error'),
        ) as (get_vlan, deferred_fn, mock_add_flow, error_log):
            self.ovs_firewall._add_flows(self.mock_br, port_with_app)
            self.assertFalse(mock_add_flow.called)
            self.assertTrue(error_log.called)

    def test_add_flows_call_tcp(self):
        port = copy.deepcopy(fake_port)
        with contextlib.nested(
            mock.patch.object(self.ovs_firewall, '_get_port_vlan',
                              return_value=100),
            mock.patch.object(self.ovs_firewall, '_get_protocol',
                              return_value=['tcp']),
            mock.patch.object(self.ovs_firewall, '_add_flow_with_range'),
            mock.patch.object(self.ovs_firewall.sg_br, 'deferred',
                              return_value=self.mock_br),
            mock.patch.object(self.mock_br, 'add_flow')
        ) as (get_vlan, get_proto, add_range_flows,
              deferred_fn, mock_add_flow):
            self.ovs_firewall._add_flows(self.mock_br, port)
            self.assertTrue(get_vlan.called)
            self.assertTrue(get_proto.called)
            self.assertTrue(add_range_flows.called)

    def test_add_flows_call_normal(self):
        port = copy.deepcopy(fake_port)
        with contextlib.nested(
            mock.patch.object(self.ovs_firewall, '_get_port_vlan',
                              return_value=100),
            mock.patch.object(self.ovs_firewall, '_get_protocol',
                              return_value=['ip']),
            mock.patch.object(self.ovs_firewall, '_add_flow_with_range'),
            mock.patch.object(self.ovs_firewall.sg_br, 'deferred',
                              return_value=self.mock_br),
            mock.patch.object(self.mock_br, 'add_flow')
        ) as (get_vlan, get_proto, add_range_flows,
              deferred_fn, mock_add_flow):
            self.ovs_firewall._add_flows(self.mock_br, port)
            self.assertTrue(get_vlan.called)
            self.assertTrue(get_proto.called)
            self.assertFalse(add_range_flows.called)
            self.assertTrue(mock_add_flow.called)

    def test_prepare_port_filter(self):
        res_port = {'security_group_source_groups': 'abc',
                    'mac_address': '00:11:22:33:44:55',
                    'network_id': "netid",
                    'id': "123",
                    'security_groups': "abc",
                    'lvid': "100",
                    'device': "123"}
        with contextlib.nested(
            mock.patch.object(self.ovs_firewall.sg_br, 'deferred',
                              return_value=self.mock_br),
            mock.patch.object(self.ovs_firewall, '_setup_aap_flows'),
            mock.patch.object(self.ovs_firewall, '_add_flows'),
            mock.patch.object(self.mock_br, 'add_flow'),
        ) as (deferred_fn, aap_flow_fn, add_flow_fn,
              mock_add_flow):
            self.ovs_firewall.prepare_port_filter(fake_port)
            aap_flow_fn.assertCalledWith(self.mock_br, fake_port)
            add_flow_fn.assertCalledWith(self.mock_br, fake_port)
            ret_port = self.ovs_firewall.filtered_ports['123']
            self.assertEqual(res_port, ret_port)

    def test_prepare_port_filter_exception(self):
        with contextlib.nested(
            mock.patch.object(self.ovs_firewall.sg_br, 'deferred',
                              return_value=self.mock_br),
            mock.patch.object(self.ovs_firewall, '_setup_aap_flows',
                              side_effect=Exception()),
            mock.patch.object(self.ovs_firewall, '_add_flows'),
            mock.patch.object(self.LOG, 'exception')
        ) as (deferred_fn, aap_flow_fn, add_flow_fn,
              exception_log):
            self.ovs_firewall.prepare_port_filter(fake_port)
            aap_flow_fn.assertCalledWith(self.mock_br, fake_port)
            self.assertFalse(add_flow_fn.called)
            self.assertTrue(exception_log.called)

    def test_remove_flows(self):
        res_port = {'security_group_source_groups': 'abc',
                    'mac_address': '00:11:22:33:44:55',
                    'network_id': "netid",
                    'id': "123",
                    'security_groups': "abc",
                    'lvid': "100",
                    'device': "123"}
        self.ovs_firewall.filtered_ports["123"] = res_port
        with contextlib.nested(
            mock.patch.object(self.ovs_firewall, '_get_port_vlan',
                              return_value=100),
            mock.patch.object(self.ovs_firewall.sg_br, 'deferred',
                              return_value=self.mock_br),
            mock.patch.object(self.mock_br, 'delete_flows')
        ) as (get_vlan, deferred_fn, mock_del_flows):
            self.ovs_firewall._remove_flows(self.mock_br, "123")
            self.assertTrue(get_vlan.called)
            self.assertEqual(5, mock_del_flows.call_count)

    def test_remove_flows_invalid_port(self):
        res_port = {'security_group_source_groups': 'abc',
                    'network_id': "netid",
                    'id': "123",
                    'security_groups': "abc",
                    'lvid': "100",
                    'device': "123"}
        self.ovs_firewall.filtered_ports["123"] = res_port
        with contextlib.nested(
            mock.patch.object(self.ovs_firewall, '_get_port_vlan',
                              return_value=100),
            mock.patch.object(self.ovs_firewall.sg_br, 'deferred',
                              return_value=self.mock_br),
            mock.patch.object(self.mock_br, 'delete_flows'),
            mock.patch.object(self.LOG, 'debug')
        ) as (get_vlan, deferred_fn, mock_del_flows, debug_log):
            self.ovs_firewall._remove_flows(self.mock_br, "123")
            self.assertTrue(get_vlan.called)
            self.assertEqual(1, mock_del_flows.call_count)
            self.assertEqual(2, debug_log.call_count)

    def test_clean_port_filters(self):
        res_port = {'security_group_source_groups': 'abc',
                    'mac_address': '00:11:22:33:44:55',
                    'network_id': "netid",
                    'id': "123",
                    'security_groups': "abc",
                    'lvid': "100",
                    'device': "123"}
        self.ovs_firewall.filtered_ports["123"] = res_port
        with contextlib.nested(
            mock.patch.object(self.ovs_firewall.sg_br, 'deferred',
                              return_value=self.mock_br),
            mock.patch.object(self.ovs_firewall, '_remove_flows')
        ) as (deferred_fn, mock_rem_flow):
            self.ovs_firewall.clean_port_filters(["123"])
            mock_rem_flow.assertCalledWith(self.mock_br, "123")
            self.assertIn("123", self.ovs_firewall.filtered_ports)

    def test_clean_port_filters_remove_port(self):
        res_port = {'security_group_source_groups': 'abc',
                    'mac_address': '00:11:22:33:44:55',
                    'network_id': "netid",
                    'id': "123",
                    'security_groups': "abc",
                    'lvid': "100",
                    'device': "123"}
        self.ovs_firewall.filtered_ports["123"] = res_port
        with contextlib.nested(
            mock.patch.object(self.ovs_firewall.sg_br, 'deferred',
                              return_value=self.mock_br),
            mock.patch.object(self.ovs_firewall, '_remove_flows')
        ) as (deferred_fn, mock_rem_flow):
            self.ovs_firewall.clean_port_filters(["123"], True)
            mock_rem_flow.assertCalledWith(self.mock_br, "123")
            self.assertNotIn("123", self.ovs_firewall.filtered_ports)

    def test_clean_port_filters_exception(self):
        res_port = {'security_group_source_groups': 'abc',
                    'mac_address': '00:11:22:33:44:55',
                    'network_id': "netid",
                    'id': "123",
                    'security_groups': "abc",
                    'lvid': "100",
                    'device': "123"}
        self.ovs_firewall.filtered_ports["123"] = res_port
        with contextlib.nested(
            mock.patch.object(self.ovs_firewall.sg_br, 'deferred',
                              return_value=self.mock_br),
            mock.patch.object(self.ovs_firewall, '_remove_flows',
                              side_effect=Exception()),
            mock.patch.object(self.LOG, 'exception')
        ) as (deferred_fn, rem_flow_fn, exception_log):
            self.ovs_firewall.clean_port_filters(["123"])
            rem_flow_fn.assertCalledWith(self.mock_br, "123")
            self.assertTrue(exception_log.called)

    def test_update_port_filters(self):
        res_port = {'security_group_source_groups': 'abc',
                    'mac_address': '00:11:22:33:44:55',
                    'network_id': "netid",
                    'id': "123",
                    'security_groups': "abc",
                    'lvid': "100",
                    'device': "123"}
        self.ovs_firewall.filtered_ports["123"] = res_port
        with contextlib.nested(
            mock.patch.object(self.ovs_firewall.sg_br, 'deferred',
                              return_value=self.mock_br),
            mock.patch.object(self.ovs_firewall, '_remove_flows'),
            mock.patch.object(self.ovs_firewall, '_setup_aap_flows'),
            mock.patch.object(self.ovs_firewall, '_add_flows'),
        ) as (deferred_fn, mock_rem_flow,
              aap_flow_fn, add_flow_fn):
            self.ovs_firewall.update_port_filter(fake_port)
            mock_rem_flow.assertCalledWith(self.mock_br, "123")
            aap_flow_fn.assertCalledWith(self.mock_br, fake_port)
            add_flow_fn.assertCalledWith(self.mock_br, fake_port)
            self.assertIn("123", self.ovs_firewall.filtered_ports)

    def test_update_port_filters_exception(self):
        res_port = {'security_group_source_groups': 'abc',
                    'mac_address': '00:11:22:33:44:55',
                    'network_id': "netid",
                    'id': "123",
                    'security_groups': "abc",
                    'lvid': "100",
                    'device': "123"}
        self.ovs_firewall.filtered_ports["123"] = res_port
        with contextlib.nested(
            mock.patch.object(self.ovs_firewall.sg_br, 'deferred',
                              return_value=self.mock_br),
            mock.patch.object(self.ovs_firewall, '_remove_flows',
                              side_effect=Exception()),
            mock.patch.object(self.ovs_firewall, '_add_flows'),
            mock.patch.object(self.LOG, 'exception')
        ) as (deferred_fn, mock_rem_flow, aap_flow_fn, exception_log):
            self.ovs_firewall.update_port_filter(fake_port)
            mock_rem_flow.assertCalledWith(self.mock_br, "123")
            self.assertFalse(aap_flow_fn.called)
            self.assertIn("123", self.ovs_firewall.filtered_ports)
            self.assertTrue(exception_log.called)
